/*
 * Name:    d_epics_motor.c 
 *
 * Purpose: MX motor driver for the EPICS APS-XFD developed motor record.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001, 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_epics.h"
#include "mx_motor.h"
#include "d_epics_motor.h"

#include "mx_clock.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_motor_record_function_list = {
	NULL,
	mxd_epics_motor_create_record_structures,
	mxd_epics_motor_finish_record_initialization,
	NULL,
	mxd_epics_motor_print_structure
};

MX_MOTOR_FUNCTION_LIST mxd_epics_motor_motor_function_list = {
	mxd_epics_motor_motor_is_busy,
	mxd_epics_motor_move_absolute,
	mxd_epics_motor_get_position,
	mxd_epics_motor_set_position,
	mxd_epics_motor_soft_abort,
	mxd_epics_motor_immediate_abort,
	mxd_epics_motor_positive_limit_hit,
	mxd_epics_motor_negative_limit_hit,
	mxd_epics_motor_find_home_position,
	mxd_epics_motor_constant_velocity_move,
	mxd_epics_motor_get_parameter,
	mxd_epics_motor_set_parameter
};

/* EPICS motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_EPICS_MOTOR_STANDARD_FIELDS
};

long mxd_epics_motor_num_record_fields
		= sizeof( mxd_epics_motor_record_field_defaults )
			/ sizeof( mxd_epics_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_motor_rfield_def_ptr
			= &mxd_epics_motor_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_epics_motor_get_pointers( MX_MOTOR *motor,
			MX_EPICS_MOTOR **epics_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_epics_motor_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( epics_motor == (MX_EPICS_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MOTOR pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_motor = (MX_EPICS_MOTOR *) motor->record->record_type_struct;

	if ( *epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_epics_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	epics_motor = (MX_EPICS_MOTOR *) malloc( sizeof(MX_EPICS_MOTOR) );

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = epics_motor;
	record->class_specific_function_list
				= &mxd_epics_motor_motor_function_list;

	motor->record = record;

	/* An EPICS motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_finish_record_initialization( MX_RECORD *record )
{
	MX_MOTOR *motor;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) record->record_class_struct;

	motor->acceleration_type = MXF_MTR_ACCEL_TIME;

	return status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_print_structure()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;
	double position, move_deadband;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type      = EPICS_MOTOR.\n\n");

	fprintf(file, "  name            = %s\n", record->name);
	fprintf(file, "  EPICS record    = %s\n",
					epics_motor->epics_record_name);

	mx_status = mx_motor_get_position( record, &position );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position        = %g %s  (%g).\n",
		motor->position, motor->units, 
		motor->raw_position.analog );
	fprintf(file, "  scale           = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf(file, "  offset          = %g %s.\n",
		motor->offset, motor->units );
	fprintf(file, "  backlash        = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf(file, "  negative limit  = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf(file, "  positive limit  = %g %s  (%g).\n",
	    motor->positive_limit, motor->units,
	    motor->raw_positive_limit.analog );

	move_deadband = motor->scale * motor->raw_move_deadband.analog;

	fprintf(file, "  move deadband   = %g %s  (%g).\n\n",
		move_deadband, motor->units,
		motor->raw_move_deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_motor_is_busy()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short done_moving_field;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->busy = FALSE;

	sprintf( pvname, "%s.DMOV", epics_motor->epics_record_name );

	mx_status = mx_caget_by_name( pvname,
				MX_CA_SHORT, 1, &done_moving_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch ( done_moving_field ) {
	case 0:
		motor->busy = TRUE;
		break;
	case 1:
		motor->busy = FALSE;
		break;
	default:
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Done moving field (.DMOV) for motor %s had an unexpected value of %d",
			epics_motor->epics_record_name, done_moving_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_move_absolute()";

	MX_EPICS_MOTOR *epics_motor;
	double new_destination;
	char pvname[60];
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	new_destination = motor->raw_destination.analog;

	sprintf( pvname, "%s.VAL", epics_motor->epics_record_name );

	mx_status = mx_caput_nowait_by_name( pvname,
					MX_CA_DOUBLE, 1, &new_destination );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_get_position()";

	MX_EPICS_MOTOR *epics_motor;
	double position;
	char pvname[60];
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s.RBV", epics_motor->epics_record_name );

	mx_status = mx_caget_by_name( pvname, MX_CA_DOUBLE, 1, &position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->raw_position.analog = position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_set_position( MX_MOTOR *motor )
{
	/* Setting the motor position is a three step process:
	 * 
	 * 1.  .SET <-- 1
	 * 2.  .VAL <-- new value
	 * 3.  .SET <-- 0
	 */

	const char fname[] = "mxd_epics_motor_set_position()";

	MX_EPICS_MOTOR *epics_motor;
	double user_set_position;
	short set_flag;
	char pvname[60];
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	user_set_position = motor->raw_set_position.analog;

	/*** Change the motor record from 'Use' to 'Set' mode. ***/

	sprintf( pvname, "%s.SET", epics_motor->epics_record_name );

	set_flag = 1;

	mx_status = mx_caput_by_name( pvname, MX_CA_SHORT, 1, &set_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*** Change the user position in the EPICS database. ***/

	sprintf( pvname, "%s.VAL", epics_motor->epics_record_name );

	mx_status = mx_caput_by_name( pvname,
				MX_CA_DOUBLE, 1, &user_set_position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*** Put the motor record back into 'Use' mode. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SET" );

	set_flag = 0;

	mx_status = mx_caput_by_name( pvname, MX_CA_SHORT, 1, &set_flag );

	return mx_status;
}

/* mxd_epics_motor_soft_abort() stops the motor in a way that
 * allows the motor to immediately be moved again without taking
 * any special effort.
 */

MX_EXPORT mx_status_type
mxd_epics_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_soft_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short stop_field;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Stop the motor. */

	sprintf( pvname, "%s.STOP", epics_motor->epics_record_name );

	stop_field = 1;		/* Stop */

	mx_status = mx_caput_by_name( pvname, MX_CA_SHORT, 1, &stop_field );

	/* The .STOP field is automatically reset to 0 by the IOC database,
	 * so the motor is immediately ready to move again.
	 */

	return mx_status;
}

/* mxd_epics_motor_immediate_abort() stops the motor in a way that
 * requires manual intervention (for example, via an MEDM screen)
 * to reset the .SMPG field back to "Go".
 */

MX_EXPORT mx_status_type
mxd_epics_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_immediate_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short smpg_field;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Stop the motor. */

	sprintf( pvname, "%s.SPMG", epics_motor->epics_record_name );

	smpg_field = 0;		/* Stop */

	mx_status = mx_caput_by_name( pvname, MX_CA_SHORT, 1, &smpg_field );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_positive_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short hls_field;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s.HLS", epics_motor->epics_record_name );

	mx_status = mx_caget_by_name( pvname, MX_CA_SHORT, 1, &hls_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( hls_field == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_negative_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short lls_field;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s.LLS", epics_motor->epics_record_name );

	mx_status = mx_caget_by_name( pvname, MX_CA_SHORT, 1, &lls_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( lls_field == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_find_home_position()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short home_value;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );

	if ( motor->home_search >= 0 ) {
		strcat( pvname, ".HOMF" );
	} else {
		strcat( pvname, ".HOMR" );
	}

	/* Send the home command without waiting for it to finish. */

	home_value = 1;

	mx_status = mx_caput_nowait_by_name( pvname,
					MX_CA_SHORT, 1, &home_value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_constant_velocity_move()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	short move_command;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->constant_velocity_move >= 0 ) {
		sprintf( pvname, "%s.JOGF", epics_motor->epics_record_name );
	} else {
		sprintf( pvname, "%s.JOGR", epics_motor->epics_record_name );
	}

	/* Send the move command without waiting for it to finish. */

	move_command = 1;

	mx_status = mx_caput_nowait_by_name( pvname,
					MX_CA_SHORT, 1, &move_command );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_get_parameter()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	float float_value;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( pvname, "%s.VELO", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->raw_speed = float_value;
		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( pvname, "%s.VBAS", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->raw_base_speed = float_value;
		break;

	case MXLV_MTR_RAW_ACCELERATION_PARAMETERS:
		sprintf( pvname, "%s.ACCL", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->raw_acceleration_parameters[0] = float_value;
		motor->raw_acceleration_parameters[1] = 0.0;
		motor->raw_acceleration_parameters[2] = 0.0;
		motor->raw_acceleration_parameters[3] = 0.0;
		break;

	case MXLV_MTR_PROPORTIONAL_GAIN:
		sprintf( pvname, "%s.PCOF", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->proportional_gain = float_value;
		break;

	case MXLV_MTR_INTEGRAL_GAIN:
		sprintf( pvname, "%s.ICOF", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->integral_gain = float_value;
		break;

	case MXLV_MTR_DERIVATIVE_GAIN:
		sprintf( pvname, "%s.DCOF", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->derivative_gain = float_value;
		break;

	case MXLV_MTR_VELOCITY_FEEDFORWARD_GAIN:
		motor->velocity_feedforward_gain = 0.0;
		break;

	case MXLV_MTR_ACCELERATION_FEEDFORWARD_GAIN:
		motor->acceleration_feedforward_gain = 0.0;
		break;

	default:
		return mx_motor_default_get_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_set_parameter()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	float float_value;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( pvname, "%s.VELO", epics_motor->epics_record_name );

		float_value = motor->raw_speed;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( pvname, "%s.VBAS", epics_motor->epics_record_name );

		float_value = motor->raw_base_speed;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;

	case MXLV_MTR_RAW_ACCELERATION_PARAMETERS:
		sprintf( pvname, "%s.ACCL", epics_motor->epics_record_name );

		float_value = motor->raw_acceleration_parameters[0];

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;

	case MXLV_MTR_AXIS_ENABLE:
		return mx_error( MXE_UNSUPPORTED, fname,
		    "Enabling or disabling EPICS motor '%s' is not supported.",
			motor->record->name);
		break;

	case MXLV_MTR_CLOSED_LOOP:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Enabling or disabling closed loop mode is not "
			"supported for EPICS motor '%s'.", motor->record->name);
		break;
	
	case MXLV_MTR_FAULT_RESET:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Fault reset is not supported for EPICS motor '%s'.",
				motor->record->name );
		break;

	case MXLV_MTR_PROPORTIONAL_GAIN:
		sprintf( pvname, "%s.PCOF", epics_motor->epics_record_name );

		float_value = motor->proportional_gain;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;

	case MXLV_MTR_INTEGRAL_GAIN:
		sprintf( pvname, "%s.ICOF", epics_motor->epics_record_name );

		float_value = motor->integral_gain;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;

	case MXLV_MTR_DERIVATIVE_GAIN:
		sprintf( pvname, "%s.DCOF", epics_motor->epics_record_name );

		float_value = motor->derivative_gain;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput_by_name( pvname,
					MX_CA_FLOAT, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;

	case MXLV_MTR_VELOCITY_FEEDFORWARD_GAIN:
		motor->velocity_feedforward_gain = 0.0;
		break;

	case MXLV_MTR_ACCELERATION_FEEDFORWARD_GAIN:
		motor->acceleration_feedforward_gain = 0.0;
		break;

	default:
		return mx_motor_default_set_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

