/*
 * Name:    d_icplus.c
 *
 * Purpose: MX interface driver for the Oxford Danfysik IC PLUS
 *          intensity monitor.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002-2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXD_ICPLUS_DEBUG	FALSE

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_ascii.h"
#include "mx_amplifier.h"
#include "mx_rs232.h"

#include "d_icplus.h"

MX_RECORD_FUNCTION_LIST mxd_icplus_record_function_list = {
	NULL,
	mxd_icplus_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxd_icplus_open,
	NULL,
	NULL,
	mxd_icplus_resynchronize
};

MX_AMPLIFIER_FUNCTION_LIST mxd_icplus_amplifier_function_list = {
	mxd_icplus_get_gain,
	mxd_icplus_set_gain,
	mxd_icplus_get_offset,
	mxd_icplus_set_offset
};

MX_RECORD_FIELD_DEFAULTS mxd_icplus_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_AMPLIFIER_STANDARD_FIELDS,
	MXD_ICPLUS_STANDARD_FIELDS
};

long mxd_icplus_num_record_fields
		= sizeof( mxd_icplus_record_field_defaults )
			/ sizeof( mxd_icplus_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_icplus_rfield_def_ptr
			= &mxd_icplus_record_field_defaults[0];

/* --- */

static mx_status_type
mxd_icplus_get_pointers( MX_RECORD *icplus_record,
			MX_AMPLIFIER **amplifier,
			MX_ICPLUS **icplus,
			const char *calling_fname )
{
	static const char fname[] = "mxd_icplus_get_pointers()";

	if ( icplus_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The RECORD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( amplifier != (MX_AMPLIFIER **) NULL ) {
		*amplifier = (MX_AMPLIFIER *)
				icplus_record->record_class_struct;

		if ( *amplifier == (MX_AMPLIFIER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_AMPLIFIER pointer for record '%s' passed by '%s' is NULL.",
				icplus_record->name, calling_fname );
		}
	}

	if ( icplus != (MX_ICPLUS **) NULL ) {
		*icplus = (MX_ICPLUS *) icplus_record->record_type_struct;

		if ( *icplus == (MX_ICPLUS *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_ICPLUS pointer for record '%s' passed by '%s' is NULL.",
				icplus_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxd_icplus_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxd_icplus_create_record_structures()";

	MX_AMPLIFIER *amplifier;
	MX_ICPLUS *icplus;

	amplifier = (MX_AMPLIFIER *) malloc( sizeof(MX_AMPLIFIER) );

	if ( amplifier == (MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate memory for an MX_AMPLIFIER structure.");
	}

	icplus = (MX_ICPLUS *) malloc( sizeof(MX_ICPLUS) );

	if ( icplus == (MX_ICPLUS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate memory for MX_ICPLUS structure." );
	}

	record->record_class_struct = amplifier;
	record->record_type_struct = icplus;

	record->class_specific_function_list
			= &mxd_icplus_amplifier_function_list;

	amplifier->record = record;
	icplus->record = record;

	icplus->discard_echoed_command_line = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_open( MX_RECORD *record )
{
	static const char fname[] = "mxd_icplus_open()";

	MX_AMPLIFIER *amplifier;
	MX_ICPLUS *icplus;
	char command[40];
	char response[80];
	int input_available;
	mx_status_type mx_status;

	mx_status = mxd_icplus_get_pointers( record,
					&amplifier, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	/* See if the serial port is configured correctly. */

	mx_status = mx_rs232_verify_configuration( icplus->rs232_record,
				9600, 8, 'N', 1, 'N', 0x0a, 0x0a );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Throw away any leftover characters. */

	mx_status = mx_rs232_discard_unwritten_output( icplus->rs232_record,
							MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_rs232_discard_unread_input( icplus->rs232_record,
							MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* See if the IC PLUS is available by trying to read
	 * the input current.
	 */

	sprintf( command, ":READ%d:CURR?", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
					response, sizeof response,
					MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* If there are still characters available from the RS-232 port,
	 * then the serial port is echoing back part of the transmitted
	 * command.  This means that the RS-232 cable is incorrectly
	 * wired, but we will attempt to continue anyway.
	 */

	mx_status = mx_rs232_input_is_available( icplus->rs232_record,
							&input_available );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( input_available ) {
		icplus->discard_echoed_command_line = TRUE;

		(void) mx_rs232_discard_unread_input( icplus->rs232_record,
						      	FALSE );

		mx_warning(
	"Some or all of the command string transmitted to IC PLUS device '%s' "
	"was echoed back to the serial port.  This means that the RS-232 "
	"cable is incorrectly wired, but we will attempt to continue by "
	"discarding the echoed characters.  However, this slows down the "
	"driver, so it would be better to fix the wiring.", record->name );

	}

	/* Set the gain, offset, and peaking time. */

	mx_status = mx_amplifier_set_gain( record, amplifier->gain );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_amplifier_set_offset( record, amplifier->offset );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_amplifier_set_time_constant( record,
						amplifier->time_constant );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_resynchronize( MX_RECORD *record )
{
	static const char fname[] = "mxd_icplus_resynchronize()";

	MX_ICPLUS *icplus;
	char command[40];
	mx_status_type mx_status;

	mx_status = mxd_icplus_get_pointers( record, NULL, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	mx_status = mx_rs232_discard_unwritten_output( icplus->rs232_record,
							MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_rs232_discard_unread_input( icplus->rs232_record,
							MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Reset the IC PLUS. */

	sprintf( command, "*RST%d", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
					NULL, 0, MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_get_gain( MX_AMPLIFIER *amplifier )
{
	static const char fname[] = "mxd_icplus_get_gain()";

	MX_ICPLUS *icplus;
	char command[40];
	char response[40];
	int exponent, num_items;
	mx_status_type mx_status;

	if ( amplifier == ( MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_AMPLIFIER pointer passed is NULL." );
	}

	mx_status = mxd_icplus_get_pointers( amplifier->record,
						NULL, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( command, ":CONF%d:CURR:RANG?", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
					response, sizeof response,
					MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &(icplus->range) );

	if ( num_items != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Unrecognizable response to a Get Input Range command '%s' "
		"by IC PLUS '%s'.  Response = '%s'.",
			command, amplifier->record->name, response );
	}

	exponent = 11 - icplus->range;

	amplifier->gain = pow( 10.0, (double) exponent );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_set_gain( MX_AMPLIFIER *amplifier )
{
	static const char fname[] = "mxd_icplus_set_gain()";

	MX_ICPLUS *icplus;
	char command[40];
	int exponent;
	mx_status_type mx_status;

	if ( amplifier == ( MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_AMPLIFIER pointer passed is NULL." );
	}

	mx_status = mxd_icplus_get_pointers( amplifier->record,
						NULL, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	exponent = (int) mx_round( log10( amplifier->gain ) );

	if ( ( exponent < 5 ) || ( exponent > 10 ) ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Requested gain %g for amplifier '%s' is outside the "
		"allowed range of 1.0e5 to 1.0e10 volts/ampere.",
			amplifier->gain, amplifier->record->name );
	}

	icplus->range = 11 - exponent;

	sprintf( command, ":CONF%d:CURR:RANG %d",
				icplus->address, icplus->range );

	mx_status = mxd_icplus_command( icplus, command,
					NULL, 0, MXD_ICPLUS_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_icplus_get_offset( MX_AMPLIFIER *amplifier )
{
	static const char fname[] = "mxd_icplus_get_offset()";

	MX_ICPLUS *icplus;
	char command[40];
	char response[40];
	int num_items, offset_percentage;
	mx_status_type mx_status;

	if ( amplifier == ( MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_AMPLIFIER pointer passed is NULL." );
	}

	mx_status = mxd_icplus_get_pointers( amplifier->record,
						NULL, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( command, ":CONF%d:CURR:OFFS?", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
					response, sizeof response,
					MXD_ICPLUS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &offset_percentage );

	if ( num_items != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Unrecognizable response to a Get Input Offset command '%s' "
		"by IC PLUS '%s'.  Response = '%s'.",
			command, amplifier->record->name, response );
	}

	amplifier->offset = 0.1 * (double) offset_percentage;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_set_offset( MX_AMPLIFIER *amplifier )
{
	static const char fname[] = "mxd_icplus_set_offset()";

	MX_ICPLUS *icplus;
	char command[40];
	long offset_percentage;
	mx_status_type mx_status;

	if ( amplifier == ( MX_AMPLIFIER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_AMPLIFIER pointer passed is NULL." );
	}

	mx_status = mxd_icplus_get_pointers( amplifier->record,
						NULL, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	offset_percentage = mx_round( 10.0 * amplifier->offset );

	if ( ( offset_percentage < 0 ) || ( offset_percentage > 99 ) ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Requested amplifier offset %g for amplifier '%s' is outside "
		"the allowed range of 0 to 9.9 volts.",
			amplifier->offset, amplifier->record->name );
	}

	sprintf( command, ":CONF%d:CURR:OFFS %ld",
				icplus->address, offset_percentage );

	mx_status = mxd_icplus_command( icplus, command,
					NULL, 0, MXD_ICPLUS_DEBUG );

	return mx_status;
}

/* === Extra functions for the use of this driver. === */

MX_EXPORT mx_status_type
mxd_icplus_command( MX_ICPLUS *icplus,
			char *command,
			char *response,
			size_t response_buffer_length,
			int debug_flag )
{
	static const char fname[] = "mxd_icplus_command()";

	char c;
	int i, max_attempts, input_available;
	unsigned long sleep_ms;
	mx_status_type mx_status, mx_status2;

	if ( icplus == (MX_ICPLUS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL MX_ICPLUS pointer passed." );
	}
	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL command buffer pointer passed." );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: sending '%s' to '%s'",
				fname, command, icplus->record->name ));
	}

	/* Send the command string. */

	mx_status = mx_rs232_putline( icplus->rs232_record, command, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The IC PLUS always sends an ACK character to acknowledge receipt
	 * of the LF terminator for the command that we just sent.  Even if
	 * we expect no other response, we must still read and discard this
	 * ACK character.
	 */

	mx_status = mx_rs232_getchar( icplus->rs232_record, &c, MXF_232_WAIT );

	if ( mx_status.code == MXE_NOT_READY ) {
		return mx_error( MXE_NOT_READY, fname,
		"No response received from IC PLUS amplifier '%s' "
		"for command '%s'.  Are you sure it is plugged in "
		"and turned on?", icplus->record->name, command );
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( c != MX_ACK ) {
		(void) mx_rs232_discard_unread_input( icplus->rs232_record,
							MXD_ICPLUS_DEBUG );

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Did not receive an ACK acknowledgement character from "
	"IC PLUS interface '%s' in response to the command '%s'.  "
	"Instead, saw a %#x (%c) character.",
		icplus->record->name, command, c, c );
	}

	/* If we expect a response, then read it in. */

	if ( response != NULL ) {
		mx_status = mx_rs232_getline( icplus->rs232_record,
					response, response_buffer_length,
					NULL, 0 );

		if ( debug_flag & (mx_status.code == MXE_SUCCESS) ) {
			MX_DEBUG(-2,("%s: received '%s' from '%s'",
				fname, response, icplus->record->name ));
		}
	} else {
		if ( debug_flag ) {
			MX_DEBUG(-2,("%s complete.", fname));
		}
	}

	/* If the IC PLUS echoes the command line back to us, then we must
	 * throw this away.
	 */

	if ( icplus->discard_echoed_command_line ) {
		max_attempts = 100;
		sleep_ms = 1;

		for ( i = 0; i < max_attempts; i++ ) {
			mx_status2 = mx_rs232_input_is_available(
							icplus->rs232_record,
							&input_available );

			if ( mx_status2.code != MXE_SUCCESS )
				break;

			if ( input_available )
				break;

			mx_msleep( sleep_ms );
		}

		if ( i >= max_attempts ) {
			mx_status = mx_error( MXE_TIMED_OUT, fname,
				"Timed out waiting for IC PLUS '%s' to echo "
				"the command line '%s' back to us.",
					icplus->record->name, command );
		}
		(void) mx_rs232_discard_unread_input(
					icplus->rs232_record, FALSE );
	}

	return mx_status;
}

