/*
 * Name:    d_icplus_aio.c
 *
 * Purpose: MX analog input and output drivers to control the Oxford Danfysik
 *          IC PLUS high voltage and read the input current.
 *
 *          The analog output driver 'icplus_voltage' controls the high
 *          voltage setting, while the analog input driver 'icplus_current'
 *          returns the ion chamber current.
 *
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXD_ICPLUS_AIO_DEBUG	FALSE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_driver.h"
#include "mx_amplifier.h"
#include "mx_analog_input.h"
#include "mx_analog_output.h"
#include "d_icplus.h"
#include "d_icplus_aio.h"

/* Initialize the ICPLUS analog I/O driver jump tables. */

MX_RECORD_FUNCTION_LIST mxd_icplus_ain_record_function_list = {
	NULL,
	mxd_icplus_ain_create_record_structures
};

MX_ANALOG_INPUT_FUNCTION_LIST mxd_icplus_ain_analog_input_function_list = {
	mxd_icplus_ain_read
};

MX_RECORD_FIELD_DEFAULTS mxd_icplus_ain_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_LONG_ANALOG_INPUT_STANDARD_FIELDS,
	MX_ANALOG_INPUT_STANDARD_FIELDS,
	MXD_ICPLUS_AINPUT_STANDARD_FIELDS
};

long mxd_icplus_ain_num_record_fields
		= sizeof( mxd_icplus_ain_record_field_defaults )
			/ sizeof( mxd_icplus_ain_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_icplus_ain_rfield_def_ptr
			= &mxd_icplus_ain_record_field_defaults[0];

/* === */

MX_RECORD_FUNCTION_LIST mxd_icplus_aout_record_function_list = {
	NULL,
	mxd_icplus_aout_create_record_structures
};

MX_ANALOG_OUTPUT_FUNCTION_LIST mxd_icplus_aout_analog_output_function_list = {
	mxd_icplus_aout_read,
	mxd_icplus_aout_write
};

MX_RECORD_FIELD_DEFAULTS mxd_icplus_aout_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_LONG_ANALOG_OUTPUT_STANDARD_FIELDS,
	MX_ANALOG_OUTPUT_STANDARD_FIELDS,
	MXD_ICPLUS_AOUTPUT_STANDARD_FIELDS
};

long mxd_icplus_aout_num_record_fields
		= sizeof( mxd_icplus_aout_record_field_defaults )
			/ sizeof( mxd_icplus_aout_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_icplus_aout_rfield_def_ptr
			= &mxd_icplus_aout_record_field_defaults[0];

static mx_status_type
mxd_icplus_ain_get_pointers( MX_ANALOG_INPUT *ainput,
			MX_ICPLUS_AINPUT **icplus_ainput,
			MX_ICPLUS **icplus,
			const char *calling_fname )
{
	static const char fname[] = "mxd_icplus_ain_get_pointers()";

	MX_RECORD *icplus_record;

	if ( ainput == (MX_ANALOG_INPUT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ANALOG_INPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( icplus_ainput == (MX_ICPLUS_AINPUT **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ICPLUS_AINPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( icplus == (MX_ICPLUS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ICPLUS pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( ainput->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_RECORD pointer for MX_ANALOG_INPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*icplus_ainput = (MX_ICPLUS_AINPUT *)
				ainput->record->record_type_struct;

	if ( *icplus_ainput == (MX_ICPLUS_AINPUT *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_ICPLUS_AINPUT pointer for record '%s' passed by '%s' is NULL.",
			ainput->record->name, calling_fname );
	}

	icplus_record = (*icplus_ainput)->icplus_record;

	if ( icplus_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_ICPLUS pointer for ICPLUS analog input "
		"record '%s' passed by '%s' is NULL.",
			ainput->record->name, calling_fname );
	}

	if ( icplus_record->mx_type != MXT_AMP_ICPLUS ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"icplus_record '%s' for IC PLUS analog input '%s' "
		"is not a IC PLUS record.  Instead, it is a '%s' record.",
			icplus_record->name, ainput->record->name,
			mx_get_driver_name( icplus_record ) );
	}

	*icplus = (MX_ICPLUS *) icplus_record->record_type_struct;

	if ( *icplus == (MX_ICPLUS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_ICPLUS pointer for IC PLUS record '%s' used by "
	"IC PLUS analog input record '%s' and passed by '%s' is NULL.",
			icplus_record->name,
			ainput->record->name,
			calling_fname );
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_icplus_aout_get_pointers( MX_ANALOG_OUTPUT *aoutput,
			MX_ICPLUS_AOUTPUT **icplus_aoutput,
			MX_ICPLUS **icplus,
			const char *calling_fname )
{
	static const char fname[] = "mxd_icplus_aout_get_pointers()";

	MX_RECORD *icplus_record;

	if ( aoutput == (MX_ANALOG_OUTPUT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ANALOG_OUTPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( icplus_aoutput == (MX_ICPLUS_AOUTPUT **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ICPLUS_AOUTPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( icplus == (MX_ICPLUS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_ICPLUS pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( aoutput->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_RECORD pointer for MX_ANALOG_OUTPUT pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*icplus_aoutput = (MX_ICPLUS_AOUTPUT *)
				aoutput->record->record_type_struct;

	if ( *icplus_aoutput == (MX_ICPLUS_AOUTPUT *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_ICPLUS_AOUTPUT pointer for record '%s' passed by '%s' is NULL.",
			aoutput->record->name, calling_fname );
	}

	icplus_record = (*icplus_aoutput)->icplus_record;

	if ( icplus_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_ICPLUS pointer for ICPLUS analog input "
		"record '%s' passed by '%s' is NULL.",
			aoutput->record->name, calling_fname );
	}

	if ( icplus_record->mx_type != MXT_AMP_ICPLUS ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"icplus_record '%s' for IC PLUS analog input '%s' "
		"is not a IC PLUS record.  Instead, it is a '%s' record.",
			icplus_record->name, aoutput->record->name,
			mx_get_driver_name( icplus_record ) );
	}

	*icplus = (MX_ICPLUS *) icplus_record->record_type_struct;

	if ( *icplus == (MX_ICPLUS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_ICPLUS pointer for IC PLUS record '%s' used by "
	"IC PLUS analog input record '%s' and passed by '%s' is NULL.",
			icplus_record->name,
			aoutput->record->name,
			calling_fname );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ===== Input functions. ===== */

MX_EXPORT mx_status_type
mxd_icplus_ain_create_record_structures( MX_RECORD *record )
{
        static const char fname[] = "mxd_icplus_ain_create_record_structures()";

        MX_ANALOG_INPUT *analog_input;
        MX_ICPLUS_AINPUT *icplus_ainput;

        /* Allocate memory for the necessary structures. */

        analog_input = (MX_ANALOG_INPUT *) malloc(sizeof(MX_ANALOG_INPUT));

        if ( analog_input == (MX_ANALOG_INPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_INPUT structure." );
        }

        icplus_ainput = (MX_ICPLUS_AINPUT *)
				malloc( sizeof(MX_ICPLUS_AINPUT) );

        if ( icplus_ainput == (MX_ICPLUS_AINPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ICPLUS_AINPUT structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_input;
        record->record_type_struct = icplus_ainput;
        record->class_specific_function_list
			= &mxd_icplus_ain_analog_input_function_list;

        analog_input->record = record;
	icplus_ainput->record = record;

	/* Raw analog input values are stored as longs. */

	analog_input->subclass = MXT_AIN_LONG;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_ain_read( MX_ANALOG_INPUT *ainput )
{
	static const char fname[] = "mxd_icplus_ain_read()";

	MX_ICPLUS_AINPUT *icplus_ainput;
	MX_ICPLUS *icplus;
	char command[80];
	char response[80];
	int num_items;
	mx_status_type mx_status;

	mx_status = mxd_icplus_ain_get_pointers( ainput,
					&icplus_ainput, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( command, ":READ%d:CURR?", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
				response, sizeof( response ),
				MXD_ICPLUS_AIO_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf(response, "%ld", &(ainput->raw_value.long_value));

	if ( num_items != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Could not find a numerical value in the response to an '%s' "
		"command to IC PLUS controller '%s' for "
		"analog input record '%s'.  Response = '%s'",
			command,
			icplus->record->name,
			ainput->record->name,
			response );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ===== Output functions. ===== */

MX_EXPORT mx_status_type
mxd_icplus_aout_create_record_structures( MX_RECORD *record )
{
        static const char fname[] = "mxd_icplus_aout_create_record_structures()";

        MX_ANALOG_OUTPUT *analog_output;
        MX_ICPLUS_AOUTPUT *icplus_aoutput;

        /* Allocate memory for the necessary structures. */

        analog_output = (MX_ANALOG_OUTPUT *)
					malloc(sizeof(MX_ANALOG_OUTPUT));

        if ( analog_output == (MX_ANALOG_OUTPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_OUTPUT structure." );
        }

        icplus_aoutput = (MX_ICPLUS_AOUTPUT *)
				malloc( sizeof(MX_ICPLUS_AOUTPUT) );

        if ( icplus_aoutput == (MX_ICPLUS_AOUTPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ICPLUS_AOUTPUT structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_output;
        record->record_type_struct = icplus_aoutput;
        record->class_specific_function_list
			= &mxd_icplus_aout_analog_output_function_list;

        analog_output->record = record;
	icplus_aoutput->record = record;

	/* Raw analog output values are stored as long. */

	analog_output->subclass = MXT_AOU_LONG;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_aout_read( MX_ANALOG_OUTPUT *aoutput )
{
	static const char fname[] = "mxd_icplus_aout_read()";

	MX_ICPLUS_AOUTPUT *icplus_aoutput;
	MX_ICPLUS *icplus;
	char command[80];
	char response[80];
	int num_items;
	mx_status_type mx_status;

	mx_status = mxd_icplus_aout_get_pointers( aoutput,
					&icplus_aoutput, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( command, ":CONF%d:VOLT?", icplus->address );

	mx_status = mxd_icplus_command( icplus, command,
				response, sizeof( response ),
				MXD_ICPLUS_AIO_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf(response, "%ld", &(aoutput->raw_value.long_value));

	if ( num_items != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Could not find a numerical value in the response to an '%s' "
		"command to IC PLUS controller '%s' for "
		"analog output record '%s'.  Response = '%s'",
			command,
			icplus->record->name,
			aoutput->record->name,
			response );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_icplus_aout_write( MX_ANALOG_OUTPUT *aoutput )
{
	static const char fname[] = "mxd_icplus_aout_write()";

	MX_ICPLUS_AOUTPUT *icplus_aoutput;
	MX_ICPLUS *icplus;
	char command[80];
	mx_status_type mx_status;

	mx_status = mxd_icplus_aout_get_pointers( aoutput,
					&icplus_aoutput, &icplus, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( command, ":CONF%d:VOLT %ld", icplus->address,
				aoutput->raw_value.long_value );

	mx_status = mxd_icplus_command( icplus, command,
					NULL, 0, MXD_ICPLUS_AIO_DEBUG );

	return mx_status;
}

