/*
 * Name:    d_network_ccd.c
 *
 * Purpose: MX network CCD driver.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_net.h"
#include "mx_ccd.h"

#include "d_network_ccd.h"

MX_RECORD_FUNCTION_LIST mxd_network_ccd_record_function_list = {
	NULL,
	mxd_network_ccd_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxd_network_ccd_open,
	NULL,
	NULL,
	NULL
};

MX_CCD_FUNCTION_LIST mxd_network_ccd_ccd_function_list = {
	mxd_network_ccd_start,
	mxd_network_ccd_stop,
	mxd_network_ccd_get_status,
	mxd_network_ccd_readout,
	mxd_network_ccd_dezinger,
	mxd_network_ccd_correct,
	mxd_network_ccd_writefile,
	mxd_network_ccd_get_parameter,
	mxd_network_ccd_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_network_ccd_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_CCD_STANDARD_FIELDS,
	MXD_NETWORK_CCD_STANDARD_FIELDS
};

long mxd_network_ccd_num_record_fields
		= sizeof( mxd_network_ccd_record_field_defaults )
			/ sizeof( mxd_network_ccd_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_ccd_rfield_def_ptr
			= &mxd_network_ccd_record_field_defaults[0];

/* --- */

static mx_status_type
mxd_network_ccd_get_pointers( MX_CCD *ccd,
			MX_NETWORK_CCD **network_ccd,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_ccd_get_pointers()";

	MX_RECORD *network_ccd_record;

	if ( ccd == (MX_CCD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_CCD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	network_ccd_record = ccd->record;

	if ( network_ccd_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for the MX_CCD pointer "
		"passed by '%s' is NULL.", calling_fname );
	}

	if ( network_ccd == (MX_NETWORK_CCD **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_NETWORK_CCD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*network_ccd = (MX_NETWORK_CCD *)
				network_ccd_record->record_type_struct;

	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxd_network_ccd_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_ccd_create_record_structures()";

	MX_CCD *ccd;
	MX_NETWORK_CCD *network_ccd;

	ccd = (MX_CCD *) malloc( sizeof(MX_CCD) );

	if ( ccd == (MX_CCD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate memory for a MX_CCD structure "
		"for record '%s'.", record->name );
	}

	network_ccd = (MX_NETWORK_CCD *) malloc( sizeof(MX_NETWORK_CCD) );

	if ( network_ccd == (MX_NETWORK_CCD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannott allocate memory for a MX_NETWORK_CCD structure"
		"for record '%s'.", record->name );
	}

	record->record_class_struct = ccd;
	record->record_type_struct = network_ccd;

	record->class_specific_function_list =
			&mxd_network_ccd_ccd_function_list;

	ccd->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ccd_open( MX_RECORD *record )
{
	const char fname[] = "mxd_network_ccd_open()";

	MX_CCD *ccd;
	MX_NETWORK_CCD *network_ccd;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	ccd = (MX_CCD *) record->record_class_struct;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.", fname, record->name));

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_network_ccd_start( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_start()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int start;
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.start", network_ccd->remote_record_name );

	start = 1;

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &start );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_stop( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_stop()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int stop;
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.stop", network_ccd->remote_record_name );

	stop = 1;

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &stop );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_get_status( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_get_status()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	unsigned long status;
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.busy", network_ccd->remote_record_name );

	mx_status = mx_get_by_name( network_ccd->server_record,
				rfname, MXFT_HEX, &status );

	ccd->status = status;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_readout( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_readout()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.readout", network_ccd->remote_record_name );

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &(ccd->readout) );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_dezinger( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_dezinger()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.dezinger", network_ccd->remote_record_name );

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &(ccd->dezinger) );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_correct( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_correct()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int correct;
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.correct", network_ccd->remote_record_name );

	correct = 1;

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &correct );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_writefile( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_writefile()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension[1];
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd, &network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( rfname, "%s.writefile_name", network_ccd->remote_record_name );

	dimension[1] = MXU_FILENAME_LENGTH;

	mx_status = mx_put_array_by_name( network_ccd->server_record,
				rfname, MXFT_STRING, 1, dimension,
				ccd->writefile_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.writefile", network_ccd->remote_record_name );

	mx_status = mx_put_by_name( network_ccd->server_record,
				rfname, MXFT_INT, &(ccd->writefile) );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ccd_get_parameter( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_get_parameter()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension[1];
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd,
						&network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,
	("%s invoked for CCD '%s', parameter type '%s' (%d)",
		fname, ccd->record->name,
		mx_get_field_label_string( ccd->record,
					ccd->parameter_type ),
		ccd->parameter_type));

	switch( ccd->parameter_type ) {
	case MXLV_CCD_DATA_FRAME_SIZE:
		sprintf( rfname, "%s.data_frame_size",
					network_ccd->remote_record_name );

		dimension[0] = 2;

		mx_status = mx_get_array_by_name( network_ccd->server_record,
			rfname, MXFT_INT, 1, dimension,
			&(ccd->data_frame_size) );

		break;
	case MXLV_CCD_BIN_SIZE:
		sprintf( rfname, "%s.bin_size",
					network_ccd->remote_record_name );

		dimension[0] = 2;

		mx_status = mx_get_array_by_name( network_ccd->server_record,
			rfname, MXFT_INT, 1, dimension,
			&(ccd->bin_size) );

		break;
	default:
		return mx_ccd_default_get_parameter_handler( ccd );
		break;
	}
	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ccd_set_parameter( MX_CCD *ccd )
{
	const char fname[] = "mxd_network_ccd_set_parameter()";

	MX_NETWORK_CCD *network_ccd;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension[1];
	mx_status_type mx_status;

	mx_status = mxd_network_ccd_get_pointers( ccd,
					&network_ccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,
	("%s invoked for CCD '%s', parameter type '%s' (%d)",
		fname, ccd->record->name,
		mx_get_field_label_string( ccd->record,
					ccd->parameter_type ),
		ccd->parameter_type));

	switch( ccd->parameter_type ) {
	case MXLV_CCD_DATA_FRAME_SIZE:
		sprintf( rfname, "%s.data_frame_size",
					network_ccd->remote_record_name );

		dimension[0] = 2;

		mx_status = mx_put_array_by_name( network_ccd->server_record,
			rfname, MXFT_INT, 1, dimension,
			&(ccd->data_frame_size) );

		break;
	case MXLV_CCD_HEADER_VARIABLE_NAME:
		sprintf( rfname, "%s.header_variable_name",
					network_ccd->remote_record_name );

		dimension[0] = MXU_CCD_HEADER_NAME_LENGTH;

		mx_status = mx_put_array_by_name( network_ccd->server_record,
			rfname, MXFT_STRING, 1, dimension,
			&(ccd->header_variable_name) );

		break;
	case MXLV_CCD_HEADER_VARIABLE_CONTENTS:
		sprintf( rfname, "%s.header_variable_contents",
					network_ccd->remote_record_name );

		dimension[0] = MXU_BUFFER_LENGTH;

		mx_status = mx_put_array_by_name( network_ccd->server_record,
			rfname, MXFT_STRING, 1, dimension,
			&(ccd->header_variable_contents) );

		break;
	default:
		return mx_ccd_default_set_parameter_handler( ccd );
		break;
	}
	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

