/*
 * Name:    d_remote_marccd.c
 *
 * Purpose: MX CCD driver for MarCCD when run in "Remote" data collection mode.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXD_REMOTE_MARCCD_DEBUG		TRUE

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <errno.h>

#include "mx_osdef.h"

#include "mx_unistd.h"
#include "mx_util.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_net.h"
#include "mx_ccd.h"

#include "d_remote_marccd.h"

MX_RECORD_FUNCTION_LIST mxd_remote_marccd_record_function_list = {
	NULL,
	mxd_remote_marccd_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxd_remote_marccd_open,
	mxd_remote_marccd_close
};

MX_CCD_FUNCTION_LIST mxd_remote_marccd_ccd_function_list = {
	mxd_remote_marccd_start,
	mxd_remote_marccd_stop,
	mxd_remote_marccd_get_status,
	mxd_remote_marccd_readout,
	mxd_remote_marccd_dezinger,
	mxd_remote_marccd_correct,
	mxd_remote_marccd_writefile,
	mxd_remote_marccd_get_parameter,
	mxd_remote_marccd_set_parameter
};

MX_RECORD_FIELD_DEFAULTS mxd_remote_marccd_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_CCD_STANDARD_FIELDS,
	MXD_REMOTE_MARCCD_STANDARD_FIELDS
};

long mxd_remote_marccd_num_record_fields
		= sizeof( mxd_remote_marccd_record_field_defaults )
			/ sizeof( mxd_remote_marccd_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_remote_marccd_rfield_def_ptr
			= &mxd_remote_marccd_record_field_defaults[0];

/* --- */

static mx_status_type
mxd_remote_marccd_get_pointers( MX_CCD *ccd,
			MX_REMOTE_MARCCD **remote_marccd,
			const char *calling_fname )
{
	static const char fname[] = "mxd_remote_marccd_get_pointers()";

	MX_RECORD *remote_marccd_record;

	if ( ccd == (MX_CCD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_CCD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	remote_marccd_record = ccd->record;

	if ( remote_marccd_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for the MX_CCD pointer "
		"passed by '%s' is NULL.", calling_fname );
	}

	if ( remote_marccd == (MX_REMOTE_MARCCD **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_REMOTE_MARCCD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*remote_marccd = (MX_REMOTE_MARCCD *)
				remote_marccd_record->record_type_struct;

	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxd_remote_marccd_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxd_remote_marccd_create_record_structures()";

	MX_CCD *ccd;
	MX_REMOTE_MARCCD *remote_marccd;

	ccd = (MX_CCD *) malloc( sizeof(MX_CCD) );

	if ( ccd == (MX_CCD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate memory for a MX_CCD structure "
		"for record '%s'.", record->name );
	}

	remote_marccd = (MX_REMOTE_MARCCD *) malloc( sizeof(MX_REMOTE_MARCCD) );

	if ( remote_marccd == (MX_REMOTE_MARCCD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannott allocate memory for a MX_REMOTE_MARCCD structure"
		"for record '%s'.", record->name );
	}

	record->record_class_struct = ccd;
	record->record_type_struct = remote_marccd;

	record->class_specific_function_list =
			&mxd_remote_marccd_ccd_function_list;

	ccd->record = record;
	remote_marccd->record = record;

	remote_marccd->marccd_socket = MX_INVALID_SOCKET;

	remote_marccd->fd_from_marccd = -1;
	remote_marccd->fd_to_marccd = -1;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_remote_marccd_find_file_descriptors( MX_REMOTE_MARCCD *remote_marccd )
{
	static const char fname[] = "mxd_remote_marccd_find_file_descriptors()";

	char *out_fd_string;
	char *in_fd_string;
	char *endptr;
	int saved_errno;

	/* If this function is called, it means that we are running on
	 * the same computer as MarCCD and that the current process was
	 * actually created by MarCCD via the "Remote" control mode.
	 *
	 * If so, then the MarCCD process will have created two pipes for us
	 * to use for communicating with MarCCD.  MarCCD will have stored
	 * the numerical values of the associated file descriptors into
	 * two environment variable:
	 *
	 *   IN_FD  - Used to read responses from MarCCD.
	 *   OUT_FD - Used to write commands to MarCCD.
	 *
	 * Thus, we must find these two environment variables and treat them as
	 * file descriptors.
	 */

	in_fd_string = getenv( "IN_FD" );

	if ( in_fd_string == (char *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
			"MarCCD environment variable 'IN_FD' was not found.  "
			"That probably means that this process was not created "
			"by MarCCD." );
	}

	out_fd_string = getenv( "OUT_FD" );

	if ( out_fd_string == (char *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
			"MarCCD environment variable 'OUT_FD' was not found.  "
			"That probably means that this process was not created "
			"by MarCCD." );
	}

	remote_marccd->fd_from_marccd =
			(int) strtoul( in_fd_string, &endptr, 10 );

	if ( *endptr != '\0' ) {
		return mx_error( MXE_IPC_IO_ERROR, fname,
		"The MarCCD created environment variable 'IN_FD' "
		"does not contain an integer.  Instead it is '%s'",
			in_fd_string );
	}

	remote_marccd->fd_to_marccd =
			(int) strtoul( out_fd_string, &endptr, 10 );

	if ( *endptr != '\0' ) {
		return mx_error( MXE_IPC_IO_ERROR, fname,
		"The MarCCD created environment variable 'OUT_FD' "
		"does not contain an integer.  Instead it is '%s'",
			out_fd_string );
	}

	remote_marccd->from_marccd =
		fdopen( remote_marccd->fd_from_marccd, "r" );

	if ( remote_marccd->from_marccd == (FILE *) NULL ) {
		saved_errno = errno;

		return mx_error( MXE_FUNCTION_FAILED, fname,
"Attempt to convert the MarCCD file descriptor into a FILE opened for "
"read failed.  Error number = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	remote_marccd->to_marccd = 
		fdopen( remote_marccd->fd_to_marccd, "w" );

	if ( remote_marccd->to_marccd == (FILE *) NULL ) {
		saved_errno = errno;

		return mx_error( MXE_FUNCTION_FAILED, fname,
"Attempt to convert the MarCCD file descriptor into a FILE opened for "
"write failed.  Error number = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_remote_marccd_connect_to_remote_server( MX_REMOTE_MARCCD *remote_marccd )
{
	static const char fname[] =
		"mxd_remote_marccd_connect_to_remote_server()";

	MX_SOCKET client_socket;
	int saved_errno;
	mx_status_type mx_status;

	/* If this function is called, it means that we are supposed to 
	 * connect via TCP/IP to 'marccd_server_socket' or some workalike
	 * running on the remote MarCCD computer.
	 */

	mx_status = mx_socket_open_as_client( &client_socket,
					remote_marccd->marccd_host,
					remote_marccd->marccd_port, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	remote_marccd->from_marccd = fdopen( client_socket, "r" );

	if ( remote_marccd->from_marccd == (FILE *) NULL ) {
		saved_errno = errno;

		return mx_error( MXE_FUNCTION_FAILED, fname,
"Attempt to convert the MarCCD socket into a FILE opened for read failed.  "
"Error number = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	remote_marccd->to_marccd = fdopen( client_socket, "w" );

	if ( remote_marccd->to_marccd == (FILE *) NULL ) {
		saved_errno = errno;

		return mx_error( MXE_FUNCTION_FAILED, fname,
"Attempt to convert the MarCCD socket into a FILE opened for write failed.  "
"Error number = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	remote_marccd->marccd_socket = client_socket;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_open( MX_RECORD *record )
{
	static const char fname[] = "mxd_remote_marccd_open()";

	MX_CCD *ccd;
	MX_REMOTE_MARCCD *remote_marccd;
	int setvbuf_status, saved_errno;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	ccd = (MX_CCD *) record->record_class_struct;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.", fname, record->name));

	if ( strlen( remote_marccd->marccd_host )  == 0 ) {

		/* If the MarCCD host name is of zero length, then we assume
		 * that the current process was started by MarCCD itself.
		 */

		mx_status =
		    mxd_remote_marccd_find_file_descriptors( remote_marccd );
	} else {
		/* Otherwise, we assume that we need to connect via TCP/IP
		 * to a server spawned by MarCCD on the MarCCD host.
		 */

		mx_status =
		    mxd_remote_marccd_connect_to_remote_server( remote_marccd );
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* It is convenient to configure the MarCCD file streams
	 * to be line buffered.
	 */

	setvbuf_status = setvbuf( remote_marccd->from_marccd,
					(char *) NULL, _IOLBF, 0 );

	if ( setvbuf_status != 0 ) {
		saved_errno = errno;

		return mx_error( MXE_IPC_IO_ERROR, fname,
		"Unable to configure reads from MarCCD as line buffered.  "
		"Errno = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	setvbuf_status = setvbuf( remote_marccd->to_marccd,
					(char *) NULL, _IOLBF, 0 );

	if ( setvbuf_status != 0 ) {
		saved_errno = errno;

		return mx_error( MXE_IPC_IO_ERROR, fname,
		"Unable to configure writes to MarCCD as line buffered.  "
		"Errno = %d, error message = '%s'",
			saved_errno, strerror( saved_errno ) );
	}

	/* Initialize the frame size. */

	mx_status = mx_ccd_get_data_frame_size( record, NULL, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_ccd_get_bin_size( record, NULL, NULL );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_close( MX_RECORD *record )
{
	static const char fname[] = "mxd_remote_marccd_close()";

	MX_CCD *ccd;
	MX_REMOTE_MARCCD *remote_marccd;
	int saved_errno, status_from_marccd, status_to_marccd;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	ccd = (MX_CCD *) record->record_class_struct;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.", fname, record->name));

	/* Tell MarCCD that it is time to exit remote mode. */

	mx_status = mxd_remote_marccd_command( remote_marccd, "end_automation",
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Wait a moment for the remote end to respond to what we have done. */

	mx_msleep(100);

	/* Shutdown the connection. */

	if ( remote_marccd->marccd_socket != MX_INVALID_SOCKET ) {

		mx_status = mx_socket_close( remote_marccd->marccd_socket );

		remote_marccd->marccd_socket = MX_INVALID_SOCKET;
	} else {
		mx_status = MX_SUCCESSFUL_RESULT;

		status_to_marccd = close( remote_marccd->fd_to_marccd );

		if ( status_to_marccd != 0 ) {
			saved_errno = errno;

			mx_status = mx_error( MXE_IPC_IO_ERROR, fname,
		"An error occurred while closing the pipe to MarCCD.  "
		"Error number = %d, error message = '%s'",
				saved_errno, strerror( saved_errno ) );
		}

		status_from_marccd = close( remote_marccd->fd_from_marccd );

		if ( status_from_marccd != 0 ) {
			saved_errno = errno;

			mx_status = mx_error( MXE_IPC_IO_ERROR, fname,
		"An error occurred while closing the pipe from MarCCD.  "
		"Error number = %d, error message = '%s'",
				saved_errno, strerror( saved_errno ) );
		}
	}

	/* (WML) Question for the day:
	 *
	 * Are the FILE structures pointed to by the from_marccd and to_marccd
	 * FILE * pointers allocated from the heap by malloc()?  If so, then
	 * the following is a memory leak.  But if not, then free()-ing them
	 * might cause a crash.  I don't currently know the answer to this.
	 */

	remote_marccd->from_marccd = NULL;
	remote_marccd->to_marccd = NULL;

	return mx_status;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_remote_marccd_start( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_start()";

	MX_REMOTE_MARCCD *remote_marccd;
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	mx_status = mxd_remote_marccd_command( remote_marccd, "start",
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_stop( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_stop()";

	MX_REMOTE_MARCCD *remote_marccd;
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	mx_status = mxd_remote_marccd_command( remote_marccd, "abort",
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_get_status( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_get_status()";

	MX_REMOTE_MARCCD *remote_marccd;
	char response[40];
	int marccd_state, num_items;
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	mx_status = mxd_remote_marccd_command( remote_marccd, "get_state",
					response, sizeof(response),
					MXD_REMOTE_MARCCD_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &marccd_state );

	if ( num_items != 1 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
"Unable to find the MarCCD status in response to a 'get_state' command.  "
"Response = '%s'", response );
	}

	switch( marccd_state ) {
	case 0:  /* IDLE */

		ccd->status = 0;
		break;
	
	case 1:  /* ACQUIRE */

		ccd->status = MXSF_CCD_ACQUIRING_FRAME;
		break;

	case 2:  /* READOUT */

		ccd->status = MXSF_CCD_READING_FRAME;
		break;

	case 3:  /* CORRECT */

		ccd->status = MXSF_CCD_CORRECTING_FRAME;
		break;

	case 4:  /* WRITING */

		ccd->status = MXSF_CCD_WRITING_FRAME;
		break;

	case 5:  /* ABORTING */

		ccd->status = MXSF_CCD_STOPPING;
		break;

	case 6:  /* UNAVAILABLE */
	case 7:  /* ERROR */
	case 8:  /* BUSY */

		ccd->status = MXSF_CCD_ERROR;
		break;
	default:
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"MarCCD returned an unrecognized state code %d for the 'get_state' command.",
			marccd_state );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_readout( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_readout()";

	MX_REMOTE_MARCCD *remote_marccd;
	char command[40];
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( command, "readout,%d", ccd->frame_type );

	mx_status = mxd_remote_marccd_command( remote_marccd, command,
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_dezinger( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_dezinger()";

	MX_REMOTE_MARCCD *remote_marccd;
	char command[40];
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( command, "dezinger,%d", ccd->frame_type );

	mx_status = mxd_remote_marccd_command( remote_marccd, command,
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_correct( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_correct()";

	MX_REMOTE_MARCCD *remote_marccd;
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	mx_status = mxd_remote_marccd_command( remote_marccd, "correct",
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_writefile( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_writefile()";

	MX_REMOTE_MARCCD *remote_marccd;
	char command[MXU_FILENAME_LENGTH + 40];
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers(ccd, &remote_marccd, fname);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for record '%s'.",
			fname, ccd->record->name));

	sprintf( command, "writefile,%s,%d",
			ccd->writefile_name, ccd->frame_type );

	mx_status = mxd_remote_marccd_command( remote_marccd, command,
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_get_parameter( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_get_parameter()";

	MX_REMOTE_MARCCD *remote_marccd;
	char response[40];
	int num_items;
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers( ccd,
						&remote_marccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,
	("%s invoked for CCD '%s', parameter type '%s' (%d)",
		fname, ccd->record->name,
		mx_get_field_label_string( ccd->record,
					ccd->parameter_type ),
		ccd->parameter_type));

	switch( ccd->parameter_type ) {
	case MXLV_CCD_DATA_FRAME_SIZE:
		mx_status = mxd_remote_marccd_command(remote_marccd, "get_size",
						response, sizeof(response),
						MXD_REMOTE_MARCCD_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%d %d",
					&(ccd->data_frame_size[0]),
					&(ccd->data_frame_size[1]) );

		if ( num_items != 2 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
			"Cannot find the MarCCD data frame size in the "
			"response to a 'get_size' command.  Response = '%s'",
				response );
		}
		break;
	case MXLV_CCD_BIN_SIZE:
		mx_status = mxd_remote_marccd_command(remote_marccd, "get_bin",
						response, sizeof(response),
						MXD_REMOTE_MARCCD_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		num_items = sscanf( response, "%d %d",
					&(ccd->bin_size[0]),
					&(ccd->bin_size[1]) );

		if ( num_items != 2 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
			"Cannot find the MarCCD bin size in the "
			"response to a 'get_bin' command.  Response = '%s'",
				response );
		}
		break;
	default:
		return mx_ccd_default_get_parameter_handler( ccd );
		break;
	}
	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_remote_marccd_set_parameter( MX_CCD *ccd )
{
	static const char fname[] = "mxd_remote_marccd_set_parameter()";

	MX_REMOTE_MARCCD *remote_marccd;
	char command[ MXU_CCD_HEADER_NAME_LENGTH + MXU_BUFFER_LENGTH + 40 ];
	mx_status_type mx_status;

	mx_status = mxd_remote_marccd_get_pointers( ccd,
						&remote_marccd, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,
	("%s invoked for CCD '%s', parameter type '%s' (%d)",
		fname, ccd->record->name,
		mx_get_field_label_string( ccd->record,
					ccd->parameter_type ),
		ccd->parameter_type));

	switch( ccd->parameter_type ) {
	case MXLV_CCD_DATA_FRAME_SIZE:
		sprintf( command, "set_bin,%d,%d", ccd->bin_size[0],
						ccd->bin_size[1] );

		mx_status = mxd_remote_marccd_command( remote_marccd, command,
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXLV_CCD_HEADER_VARIABLE_NAME:
		/* Nothing needs to be done beyond storing the name
		 * for later use by a write to 'header_variable_contents'.
		 */

		break;
	case MXLV_CCD_HEADER_VARIABLE_CONTENTS:
		sprintf( command, "header,%s=%s", ccd->header_variable_name,
					ccd->header_variable_contents );

		mx_status = mxd_remote_marccd_command( remote_marccd, command,
					NULL, 0, MXD_REMOTE_MARCCD_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	default:
		return mx_ccd_default_set_parameter_handler( ccd );
		break;
	}
	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* --- */

MX_EXPORT mx_status_type
mxd_remote_marccd_command( MX_REMOTE_MARCCD *remote_marccd, char *command,
				char *response, size_t max_response_length,
				int flags )
{
	static const char fname[] = "mxd_remote_marccd_command()";

	size_t length;

	if ( remote_marccd == (MX_REMOTE_MARCCD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_REMOTE_MARCCD pointer passed was NULL." );
	}
	if ( command == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The command string pointer passed was NULL." );
	}

	if ( remote_marccd->to_marccd == (FILE *) NULL ) {
		return mx_error( MXE_IPC_IO_ERROR, fname,
	"The connection to MarCCD has not been correctly initialized." );
	}

	if ( remote_marccd->from_marccd == (FILE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The connection to MarCCD has not been fully initialized." );
	}

	if ( flags ) {
		MX_DEBUG(-2,("%s: sent '%s' to '%s'",
			fname, command, remote_marccd->record->name));
	}

	/* WML: The following has not been tested yet and is probably
	 * overoptimistic.
	 */

	fprintf( remote_marccd->to_marccd, "%s\n", command );

	if ( response != NULL ) {
		fgets( response, max_response_length,
			remote_marccd->from_marccd );

		/* Delete any trailing newline if present. */

		length = strlen( response );

		if ( response[length-1] == '\n' )
			response[length-1] = '\0';

		if ( flags ) {
			MX_DEBUG(-2,("%s: received '%s' from '%s'",
				fname, response, remote_marccd->record->name ));
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

