/*
 * Name:    i_aps_adcmod2.c
 *
 * Purpose: MX interface driver for the ADCMOD2 electrometer system designed
 *          by Steve Ross of the Advanced Photon Source.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_vme.h"

#include "i_aps_adcmod2.h"

MX_RECORD_FUNCTION_LIST mxi_aps_adcmod2_record_function_list = {
	NULL,
	mxi_aps_adcmod2_create_record_structures,
	NULL,
	NULL,
	mxi_aps_adcmod2_print_structure,
	NULL,
	NULL,
	mxi_aps_adcmod2_open
};

MX_RECORD_FIELD_DEFAULTS mxi_aps_adcmod2_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_APS_ADCMOD2_STANDARD_FIELDS
};

long mxi_aps_adcmod2_num_record_fields
		= sizeof( mxi_aps_adcmod2_record_field_defaults )
			/ sizeof( mxi_aps_adcmod2_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_aps_adcmod2_rfield_def_ptr
			= &mxi_aps_adcmod2_record_field_defaults[0];

/* --- */

static mx_status_type
mxi_aps_adcmod2_get_pointers( MX_RECORD *record,
			MX_APS_ADCMOD2 **aps_adcmod2,
			const char *calling_fname )
{
	const char fname[] = "mxi_aps_adcmod2_get_pointers()";

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_RECORD pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( aps_adcmod2 == (MX_APS_ADCMOD2 **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_APS_ADCMOD2 pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*aps_adcmod2 = (MX_APS_ADCMOD2 *) (record->record_type_struct);

	if ( *aps_adcmod2 == (MX_APS_ADCMOD2 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_APS_ADCMOD2 pointer for record '%s' is NULL.",
			record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/*=== Public functions ===*/

MX_EXPORT mx_status_type
mxi_aps_adcmod2_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_aps_adcmod2_create_record_structures()";

	MX_APS_ADCMOD2 *aps_adcmod2;
	int i;

	/* Allocate memory for the necessary structures. */

	aps_adcmod2 = (MX_APS_ADCMOD2 *) malloc( sizeof(MX_APS_ADCMOD2) );

	if ( aps_adcmod2 == (MX_APS_ADCMOD2 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_APS_ADCMOD2 structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;
	record->record_type_struct = aps_adcmod2;
	record->class_specific_function_list = NULL;

	aps_adcmod2->record = record;

	for ( i = 0; i < MX_APS_ADCMOD2_NUM_AMPLIFIERS; i++ ) {
		aps_adcmod2->amplifier_array[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_aps_adcmod2_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxi_aps_adcmod2_print_structure()";

	MX_APS_ADCMOD2 *aps_adcmod2;
	MX_RECORD *this_record;
	int i;
	mx_status_type mx_status;

	mx_status = mxi_aps_adcmod2_get_pointers( record, &aps_adcmod2, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "Record '%s':\n\n", record->name);
	fprintf(file, "  name                       = \"%s\"\n",
						record->name);

	fprintf(file, "  superclass                 = interface\n");
	fprintf(file, "  class                      = generic\n");
	fprintf(file, "  type                       = APS_ADCMOD2\n");
	fprintf(file, "  label                      = \"%s\"\n",
						record->label);

	fprintf(file, "  vme record                 = \"%s\"\n",
						aps_adcmod2->vme_record->name);

	fprintf(file, "  crate_number               = %lu\n",
						aps_adcmod2->crate_number);

	fprintf(file, "  base address               = %#lx\n",
						aps_adcmod2->base_address);

	fprintf(file, "  amplifiers in use          = (");

	for ( i = 0; i < MX_APS_ADCMOD2_NUM_AMPLIFIERS; i++ ) {

		this_record = aps_adcmod2->amplifier_array[i];

		if ( this_record == NULL ) {
			fprintf( file, "NULL" );
		} else {
			fprintf( file, "%s", this_record->name );
		}
		if ( i != (MX_APS_ADCMOD2_NUM_AMPLIFIERS - 1) ) {
			fprintf( file, ", " );
		}
	}

	fprintf(file, ")\n");

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_aps_adcmod2_open( MX_RECORD *record )
{
	const char fname[] = "mxi_aps_adcmod2_open()";

	MX_APS_ADCMOD2 *aps_adcmod2;
	mx_status_type mx_status;

	mx_status = mxi_aps_adcmod2_get_pointers( record, &aps_adcmod2, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return mx_status;
}

/* ====== */

MX_EXPORT mx_status_type
mxi_aps_adcmod2_in16( MX_APS_ADCMOD2 *aps_adcmod2, unsigned long offset,
				mx_uint16_type *word_value )
{
	const char fname[] = "mxi_aps_adcmod2_in16()";

	unsigned long address;
	mx_status_type mx_status;

	if ( aps_adcmod2 == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"NULL aps_adcmod2 pointer passed." );
	}
	if ( offset > 0xffffff ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal VME address offset %lu for APS_ADCMOD2 record '%s'",
				offset, aps_adcmod2->record->name );
	}

	address = aps_adcmod2->base_address + offset;

	mx_status = mx_vme_in16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, word_value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_aps_adcmod2_out16( MX_APS_ADCMOD2 *aps_adcmod2, unsigned long offset,
				mx_uint16_type word_value )
{
	const char fname[] = "mxi_aps_adcmod2_out16()";

	unsigned long address;
	mx_status_type mx_status;

	if ( aps_adcmod2 == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"NULL aps_adcmod2 pointer passed." );
	}
	if ( offset > 0xffffff ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal VME address offset %lu for APS_ADCMOD2 record '%s'",
				offset, aps_adcmod2->record->name );
	}

	address = aps_adcmod2->base_address + offset;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, word_value );

	return mx_status;
}

#if 0

MX_EXPORT mx_status_type
mxi_aps_adcmod2_command( MX_APS_ADCMOD2 *aps_adcmod2, int electrometer_number,
				mx_uint16_type command, mx_uint16_type value )
{
	unsigned long address;
	mx_status_type mx_status;

	/* Write the device type which is always 0xA000 for this device. */

	address = aps_adcmod2->base_address + 0x18;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, 0xA000 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Write the command number. */

	address = aps_adcmod2->base_address + 0x10;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, command );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Write the command value. */

	address = aps_adcmod2->base_address + 0x08;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the command. */

	address = aps_adcmod2->base_address + 0x20;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, 0xCCCC );

	return mx_status;
}

#else

MX_EXPORT mx_status_type
mxi_aps_adcmod2_command( MX_APS_ADCMOD2 *aps_adcmod2, int electrometer_number,
				mx_uint16_type command, mx_uint16_type value )
{
	unsigned long address;
	mx_status_type mx_status;

	/* Write the command value. */

	address = aps_adcmod2->base_address + 0x08;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Write the command number. */

	address = aps_adcmod2->base_address + 0x10;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, command );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Write the device type which is always 0xA000 for this device. */

	address = aps_adcmod2->base_address + 0x18;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, 0xA000 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the command. */

	address = aps_adcmod2->base_address + 0x20;

	mx_status = mx_vme_out16( aps_adcmod2->vme_record,
				aps_adcmod2->crate_number, MXF_VME_A24,
				address, 0xCCCC );

	return mx_status;
}

#endif

