/*
 * Name:    i_xia_handel.c
 *
 * Purpose: MX driver for the X-Ray Instrumentation Associates Handel library
 *          used by the DXP series of multichannel analyzer.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXI_XIA_HANDEL_DEBUG	FALSE

#include <stdio.h>
#include <limits.h>

#include "mxconfig.h"

#if HAVE_XIA_HANDEL

#include <stdlib.h>
#include <errno.h>

#include <handel.h>
#include <handel_errors.h>
#include <handel_generic.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_types.h"
#include "mx_alloc.h"
#include "mx_mca.h"
#include "i_xia_handel.h"
#include "d_xia_dxp_mca.h"
#include "u_xia_dxp.h"

MX_RECORD_FUNCTION_LIST mxi_xia_handel_record_function_list = {
	NULL,
	mxi_xia_handel_create_record_structures,
	mxi_xia_handel_finish_record_initialization,
	NULL,
	NULL,
	NULL,
	NULL,
	mxi_xia_handel_open,
	NULL,
	NULL,
	mxi_xia_handel_resynchronize,
	mxi_xia_handel_special_processing_setup
};

MX_RECORD_FIELD_DEFAULTS mxi_xia_handel_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_XIA_HANDEL_STANDARD_FIELDS
};

long mxi_xia_handel_num_record_fields
		= sizeof( mxi_xia_handel_record_field_defaults )
			/ sizeof( mxi_xia_handel_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_xia_handel_rfield_def_ptr
			= &mxi_xia_handel_record_field_defaults[0];

static void mxi_xia_handel_set_data_available_flags(
					MX_XIA_HANDEL *xia_handel,
					int flag_value );

static mx_status_type mxi_xia_handel_process_function( void *record_ptr,
						void *record_field_ptr,
						int operation );

extern const char *mxi_xia_handel_strerror( int xia_status );

static mx_status_type
mxi_xia_handel_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			MX_XIA_HANDEL **xia_handel,
			const char *calling_fname )
{
	static const char fname[] = "mxi_xia_handel_get_pointers()";

	MX_RECORD *xia_dxp_record;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_handel == (MX_XIA_HANDEL **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_HANDEL pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	xia_dxp_record = (*xia_dxp_mca)->xia_dxp_record;

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for MCA record '%s' is NULL.",
			mca->record->name );
	}

	*xia_handel = (MX_XIA_HANDEL *) xia_dxp_record->record_type_struct;

	if ( (*xia_handel) == (MX_XIA_HANDEL *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_HANDEL pointer for XIA DXP record '%s' is NULL.",
			xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_xia_handel_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxi_xia_handel_create_record_structures()";

	MX_XIA_HANDEL *xia_handel;

	/* Allocate memory for the necessary structures. */

	xia_handel = (MX_XIA_HANDEL *) malloc( sizeof(MX_XIA_HANDEL) );

	if ( xia_handel == (MX_XIA_HANDEL *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_XIA_HANDEL structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;;
	record->record_type_struct = xia_handel;

	record->class_specific_function_list = NULL;

	xia_handel->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_finish_record_initialization( MX_RECORD *record )
{
	static const char fname[] = "mxi_xia_handel_finish_record_initialization()";

	MX_XIA_HANDEL *xia_handel;

	xia_handel = (MX_XIA_HANDEL *) record->record_type_struct;

	if ( xia_handel == (MX_XIA_HANDEL *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_HANDEL pointer for record '%s' is NULL.",
			record->name );
	}

	/* We cannot allocate the mca_record_array structures until
	 * mxi_xia_handel_open() is invoked, so for now we initialize
	 * these data structures to all zeros.
	 */

	xia_handel->num_mcas = 0;
	xia_handel->mca_record_array = NULL;

	return MX_SUCCESSFUL_RESULT;
}

/* mxi_xia_handel_set_data_available_flags() sets mca->new_data_available
 * and xia_dxp_mca->new_statistics_available for all of the MCAs controlled
 * by this Handel interface.
 */

static void
mxi_xia_handel_set_data_available_flags( MX_XIA_HANDEL *xia_handel,
					int flag_value )
{
	MX_RECORD *mca_record;
	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	unsigned long i;

	for ( i = 0; i < xia_handel->num_mcas; i++ ) {

		mca_record = (xia_handel->mca_record_array)[i];

		mca = (MX_MCA *) mca_record->record_class_struct;

		xia_dxp_mca = (MX_XIA_DXP_MCA *) mca_record->record_type_struct;

		mca->new_data_available = flag_value;
		xia_dxp_mca->new_statistics_available = flag_value;
	}
}

/*
 * mxi_xia_handel_load_config() encapsulates all of the operations that
 * must happen when a new configuration file is loaded.
 */

static mx_status_type
mxi_xia_handel_load_config( MX_XIA_HANDEL *xia_handel )
{
	static const char fname[] = "mxi_xia_handel_load_config()";

	MX_RECORD *mca_record;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	int i, xia_status;

	/* Load the configuration file. */

	MX_DEBUG(-4,("%s: Loading Handel configuration '%s'.",
		fname, xia_handel->config_filename));

	xia_status = xiaLoadSystem( "handel_ini", xia_handel->config_filename );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to open the Handel detector configuration file '%s'.  "
		"Error code = %d, '%s'",
			xia_handel->config_filename,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	/* Start the system. */

	xia_status = xiaStartSystem();

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to start the MCA using Handel configuration file '%s'."
		"  Error code = %d, '%s'",
			xia_handel->config_filename,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	/* Invalidate the old presets for all of the MCAs by setting the
	 * old preset type to an illegal value.
	 */

	if ( xia_handel->mca_record_array != NULL ) {
		for ( i = 0; i < xia_handel->num_mcas; i++ ) {
			mca_record = xia_handel->mca_record_array[i];

			if ( mca_record != NULL ) {
				xia_dxp_mca = (MX_XIA_DXP_MCA *)
					mca_record->record_type_struct;

				xia_dxp_mca->old_preset_type
					= (mx_uint32_type) ULONG_MAX;
			}
		}
	}


	MX_DEBUG(-4,("%s: Successfully loaded Handel configuration '%s'.",
		fname, xia_handel->config_filename));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_open( MX_RECORD *record )
{
	static const char fname[] = "mxi_xia_handel_open()";

	MX_XIA_HANDEL *xia_handel;
	int xia_status;
	unsigned int i, num_mcas;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_handel = (MX_XIA_HANDEL *) (record->record_type_struct);

	if ( xia_handel == (MX_XIA_HANDEL *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_HANDEL pointer for record '%s' is NULL.", record->name);
	}

	/* Redirect Handel and Xerxes output to the MX output functions. */

	mx_status = mxu_xia_dxp_replace_output_functions();

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Initialize mca_record_array to NULL, so that load_config
	 * below knows not to try to change the MCA presets.
	 */

	xia_handel->mca_record_array = NULL;

	/* Initialize Handel. */

	xia_status = xiaInitHandel();

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to initialize Handel.  "
		"Error code = %d, '%s'",
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	/* Load the detector configuration. */

	mx_status = mxi_xia_handel_load_config( xia_handel );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Find out how many detector channels (MCAs) there are. */

	xia_status = xiaGetNumDetectors( &num_mcas );

	MX_DEBUG(-2,("%s: num_mcas = %d", fname, num_mcas));

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to find out how many detector channels are in "
		"the current configuration '%s'.  "
		"Error code = %d, '%s'",
			xia_handel->config_filename,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	/* Allocate and initialize an array to store pointers to all of
	 * the MCA records used by this interface.
	 */

	xia_handel->mca_record_array = ( MX_RECORD ** ) malloc(
					num_mcas * sizeof( MX_RECORD * ) );

	if ( xia_handel->mca_record_array == ( MX_RECORD ** ) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %d element array "
		"of MX_RECORD pointers for the mca_record_array data "
		"structure of record '%s'.", num_mcas, record->name );
	}

	for ( i = 0; i < num_mcas; i++ ) {
		xia_handel->mca_record_array[i] = NULL;
	}

	xia_handel->num_mcas = num_mcas;

#if 1
	{
		char **detector_aliases;
		long dimension[2];
		size_t element_size[2];

		/* Display the list of detectors. */

		dimension[0] = xia_handel->num_mcas;
		dimension[1] = MAXALIAS_LEN;

		element_size[0] = sizeof(char);
		element_size[1] = sizeof(char *);

		detector_aliases = (char **)
			mx_allocate_array( 2, dimension, element_size );

		if ( detector_aliases == (char **) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate a %ld by %ld string array.",
				dimension[0], dimension[1] );
		}

		xia_status = xiaGetDetectors( detector_aliases );

		if ( xia_status != XIA_SUCCESS ) {
			return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
			"Unable to get the list of detector aliases for "
			"configuration file '%s'.  "
			"Error code = %d, '%s'",
			xia_handel->config_filename,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
		}

		for ( i = 0; i < xia_handel->num_mcas; i++ ) {
			MX_DEBUG(-2,("%s: detector(%d) = '%s'",
			fname, i, detector_aliases[i]));
		}

		mx_free_array( detector_aliases, 2, dimension, element_size );
	}
#endif
#if 1
	{
		char **module_aliases;
		long dimension[2];
		size_t element_size[2];

		/* Display the list of modules. */

		dimension[0] = xia_handel->num_mcas;
		dimension[1] = MAXALIAS_LEN;

		element_size[0] = sizeof(char);
		element_size[1] = sizeof(char *);

		module_aliases = (char **)
			mx_allocate_array( 2, dimension, element_size );

		if ( module_aliases == (char **) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate a %ld by %ld string array.",
				dimension[0], dimension[1] );
		}

		xia_status = xiaGetModules( module_aliases );

		if ( xia_status != XIA_SUCCESS ) {
			return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
			"Unable to get the list of module aliases for "
			"configuration file '%s'.  "
			"Error code = %d, '%s'",
			xia_handel->config_filename,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
		}

		for ( i = 0; i < xia_handel->num_mcas; i++ ) {
			MX_DEBUG(-2,("%s: module(%d) = '%s'",
			fname, i, module_aliases[i]));
		}

		mx_free_array( module_aliases, 2, dimension, element_size );
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_resynchronize( MX_RECORD *record )
{
	static const char fname[] = "mxi_xia_handel_resynchronize()";

	MX_XIA_HANDEL *xia_handel;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_handel = (MX_XIA_HANDEL *) (record->record_type_struct);

	if ( xia_handel == (MX_XIA_HANDEL *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_HANDEL pointer for record '%s' is NULL.", record->name);
	}

	mx_status = mxi_xia_handel_load_config( xia_handel );

	return mx_status;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_handel_is_busy( MX_MCA *mca,
			int *busy_flag,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_is_busy()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	unsigned long run_active, mask;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_status = xiaGetRunData( xia_dxp_mca->system_channel,
					"run_active", &run_active );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot get DXP running status for MCA '%s'.  "
		"Error code = %d, '%s'",
			mca->record->name, xia_status,
			mxi_xia_handel_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: run_active = %#lx", fname, run_active));
	}

#if 0
	mask = XIA_RUN_HARDWARE | XIA_RUN_HANDEL | XIA_RUN_CT;
#else
	mask = XIA_RUN_HARDWARE;
#endif

	if ( run_active & mask ) {
		*busy_flag = TRUE;
	} else {
		*busy_flag = FALSE;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_read_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type *value_ptr,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_read_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s'.",
			fname, mca->record->name,
			xia_dxp_mca->system_channel, parameter_name));
	}

	xia_status = xiaGetParameter( xia_dxp_mca->system_channel,
						parameter_name, &short_value );
	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read DXP parameter '%s' for system channel %d "
		"of XIA DXP interface '%s'.  "
		"Error code = %d, '%s'",
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_handel_strerror( xia_status ) );
	}

	*value_ptr = (mx_uint32_type) short_value;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: value = %lu", fname, *value_ptr));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_write_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_write_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s', value = %lu.",
			fname, mca->record->name, xia_dxp_mca->system_channel,
			parameter_name, (unsigned long) value));
	}

	/* Make sure the run is stopped before trying to change a parameter. */

	xia_status = xiaStopRun( xia_dxp_mca->system_channel );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot stop data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	/* Send the parameter value. */

	short_value = (unsigned short) value;

	xia_status = xiaSetParameter( xia_dxp_mca->system_channel,
						parameter_name, short_value );
	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot write the value %hu to DXP parameter '%s' "
		"for system channel %d of XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", short_value,
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_handel_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write to parameter '%s' succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_write_parameter_to_all_channels( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_write_parameter_to_all_channels()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', parameter '%s', value = %lu.",
			fname, mca->record->name,
			parameter_name, (unsigned long) value));
	}

	/* Make sure the run is stopped before trying to change a parameter. */

	xia_status = xiaStopRun( xia_dxp_mca->system_channel );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot stop data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	short_value = (unsigned short) value;

#if 0
	/* I don't know yet whether there is a Handel function for this. */

#else
	/* A temporary kludge which does work. */

	{
		MX_RECORD *mca_record;
		MX_XIA_DXP_MCA *local_dxp_mca;
		unsigned long i;

		for ( i = 0; i < xia_handel->num_mcas; i++ ) {

			mca_record = xia_handel->mca_record_array[i];

			if ( mca_record == (MX_RECORD *) NULL ) {

				continue;    /* Skip this one. */
			}

			local_dxp_mca = (MX_XIA_DXP_MCA *)
					mca_record->record_type_struct;

			MX_DEBUG(-2,("%s: writing to system channel %d.",
				fname, local_dxp_mca->system_channel));

			xia_status = xiaSetParameter(
					local_dxp_mca->system_channel,
					parameter_name, short_value );

			if ( xia_status != XIA_SUCCESS ) {
				return mx_error( MXE_DEVICE_ACTION_FAILED,fname,
			"Cannot write the value %hu to DXP parameter '%s' "
			"for system channel %d of XIA DXP interface '%s'.  "
			"Error code = %d, '%s'", short_value,
				parameter_name, local_dxp_mca->system_channel,
				mca->record->name, xia_status,
				mxi_xia_handel_strerror( xia_status ) );
			}
		}
	}
#endif
	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write parameter '%s' to all channels succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_start_run( MX_MCA *mca,
			int clear_flag,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_start_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	unsigned short resume_flag;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', clear_flag = %d.",
				fname, mca->record->name, clear_flag));
	}

	mxi_xia_handel_set_data_available_flags( xia_handel, TRUE );

	if ( clear_flag ) {
		resume_flag = 0;
	} else {
		resume_flag = 1;
	}

	xia_status = xiaStartRun( xia_dxp_mca->system_channel, resume_flag );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot start data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}


MX_EXPORT mx_status_type
mxi_xia_handel_stop_run( MX_MCA *mca,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_stop_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s'.", fname, mca->record->name));
	}

	xia_status = xiaStopRun( xia_dxp_mca->system_channel );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot stop data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_read_spectrum( MX_MCA *mca,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_read_spectrum()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	int xia_status;
	unsigned long *array_ptr;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s: xia_dxp_mca->use_mca_channel_array = %d",
		fname, xia_dxp_mca->use_mca_channel_array));

	if ( xia_dxp_mca->use_mca_channel_array ) {
		array_ptr = mca->channel_array;
	} else {
		array_ptr = xia_dxp_mca->spectrum_array;
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: reading out %ld channels from MCA '%s'.",
			fname, mca->current_num_channels, mca->record->name));
	}

	xia_status = xiaGetRunData( xia_dxp_mca->system_channel,
					"mca", array_ptr );

	if ( xia_status != XIA_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read the spectrum for XIA MCA '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_handel_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: readout from MCA '%s' complete.",
			fname, mca->record->name));
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxi_xia_handel_get_time( MX_MCA *mca, MX_XIA_DXP_MCA *xia_dxp_mca,
			char *name, double *time, int debug_flag )
{
	char parameter_name[80];
	double local_clock_ticks;
	mx_uint32_type value;
	mx_status_type mx_status;

	local_clock_ticks = 0.0;

	/* xxxxTIME2 */

	sprintf( parameter_name, "%s2", name );

	mx_status = mxi_xia_handel_read_parameter( mca, parameter_name,
						&value, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	local_clock_ticks = (double) value;

	/* xxxxTIME0 */

	local_clock_ticks *= 65536.0;

	sprintf( parameter_name, "%s0", name );

	mx_status = mxi_xia_handel_read_parameter( mca, parameter_name,
						&value, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	local_clock_ticks += (double) value;

	/* xxxxTIME1 */

	local_clock_ticks *= 65536.0;

	sprintf( parameter_name, "%s1", name );

	mx_status = mxi_xia_handel_read_parameter( mca, parameter_name,
						&value, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	local_clock_ticks += (double) value;

	/*---*/

	*time = local_clock_ticks * xia_dxp_mca->clock_tick;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxi_xia_handel_get_32bit_value( MX_MCA *mca, MX_XIA_DXP_MCA *xia_dxp_mca,
				char *name, mx_uint32_type *value,
				int debug_flag )
{
	char parameter_name[80];
	mx_uint32_type local_value;
	int i;
	mx_status_type mx_status;

	(*value) = 0L;

	for ( i = 0; i <=1 ; i++ ) {
		(*value) *= 65536L;

		sprintf( parameter_name, "%s%d", name, i );

		mx_status = mxi_xia_handel_read_parameter( mca, parameter_name,
						&local_value, debug_flag );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		(*value) += local_value;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_read_statistics( MX_MCA *mca,
			int debug_flag )
{
	static const char fname[] = "mxi_xia_handel_read_statistics()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_HANDEL *xia_handel;
	mx_status_type mx_status;

	mx_status = mxi_xia_handel_get_pointers( mca,
					&xia_dxp_mca, &xia_handel, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_time( mca, xia_dxp_mca, "LIVETIME",
					&(mca->live_time), debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_time( mca, xia_dxp_mca, "REALTIME",
					&(mca->real_time), debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_32bit_value( mca, xia_dxp_mca,
				"FASTPEAKS", &(xia_dxp_mca->num_fast_peaks),
				debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_dxp_mca->input_count_rate = mx_divide_safely(
					(double) xia_dxp_mca->num_fast_peaks,
						mca->live_time );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_32bit_value( mca, xia_dxp_mca,
				"EVTSINRUN", &(xia_dxp_mca->num_events),
				debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_dxp_mca->output_count_rate = mx_divide_safely(
					(double) xia_dxp_mca->num_events,
						mca->real_time );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_32bit_value( mca, xia_dxp_mca,
				"UNDRFLOWS", &(xia_dxp_mca->num_underflows),
				debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_handel_get_32bit_value( mca, xia_dxp_mca,
				"OVERFLOWS", &(xia_dxp_mca->num_overflows),
				debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,(
	"%s: Record '%s', channel %d, live_time = %g, real_time = %g, "
	"icr = %g, ocr = %g, nevent = %lu", fname,
			mca->record->name,
			xia_dxp_mca->system_channel,
			mca->live_time,
			mca->real_time,
			xia_dxp_mca->input_count_rate,
			xia_dxp_mca->output_count_rate,
			xia_dxp_mca->num_events ));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_handel_special_processing_setup( MX_RECORD *record )
{
	static const char fname[] = "mxi_xia_handel_special_processing_setup()";

	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	record_field_array = record->record_field_array;

	for ( i = 0; i < record->num_record_fields; i++ ) {

		record_field = &record_field_array[i];

		switch( record_field->label_value ) {
		case MXLV_XIA_HANDEL_CONFIG_FILENAME:
			record_field->process_function
					    = mxi_xia_handel_process_function;
			break;
		default:
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

const char *
mxi_xia_handel_strerror( int handel_status )
{
	static struct {
		int handel_error_code;
		const char handel_error_message[140];
	} error_code_table[] = {
		{   0, "XIA_SUCCESS" },
		{   1, "XIA_OPEN_FILE" },
		{   2, "XIA_FILEERR" },
		{   3, "XIA_NOSECTION" },
		{   4, "XIA_FORMAT_ERROR" },
		{   5, "XIA_ILLEGAL_OPERATION - "
			"Attempted to configure options in an illegal order" },
		{   6, "XIA_FILE_RA - "
			"File random access unable to find name-value pair" },
		{ 201, "XIA_UNKNOWN_DECIMATION - "
	"The decimation read from the hardware does not match a known value" },
		{ 202, "XIA_SLOWLEN_OOR - "
			"Calculated SLOWLEN value is out-of-range" },
		{ 203, "XIA_SLOWGAP_OOR - "
			"Calculated SLOWGAP value is out-of-range" },
		{ 204, "XIA_SLOWFILTER_OOR - "
			"Attempt to set the Peaking or Gap time s.t. P+G>31" },
		{ 205, "XIA_FASTLEN_OOR - "
			"Calculated FASTLEN value is out-of-range" },
		{ 206, "XIA_FASTGAP_OOR - "
			"Calculated FASTGAP value is out-of-range" },
		{ 207, "XIA_FASTFILTER_OOR - "
			"Attempt to set the Peaking or Gap time s.t. P+G>31" },
		{ 301, "XIA_INITIALIZE" },
		{ 302, "XIA_NO_ALIAS" },
		{ 303, "XIA_ALIAS_EXISTS" },
		{ 304, "XIA_BAD_VALUE" },
		{ 305, "XIA_INFINITE_LOOP" },
		{ 306, "XIA_BAD_NAME - Specified name is not valid" },
		{ 307, "XIA_BAD_PTRR - "
			"Specified PTRR is not valid for this alias" },
		{ 308, "XIA_ALIAS_SIZE - Alias name has too many characters" },
		{ 309, "XIA_NO_MODULE - "
			"Must define at least one module before" },
		{ 310, "XIA_BAD_INTERFACE - "
			"The specified interface does not exist" },
		{ 311, "XIA_NO_INTERFACE - "
	"An interface must be defined before more information is specified" },
		{ 312, "XIA_WRONG_INTERFACE - "
	"Specified information doesn't apply to this interface" },
		{ 313, "XIA_NO_CHANNELS - "
			"Number of channels for this module is set to 0" },
		{ 314, "XIA_BAD_CHANNEL - "
			"Specified channel index is invalid or out-of-range" },
		{ 315, "XIA_NO_MODIFY - "
			"Specified name cannot be modified once set" },
		{ 316, "XIA_INVALID_DETCHAN - "
			"Specified detChan value is invalid" },
		{ 317, "XIA_BAD_TYPE - "
			"The DetChanElement type specified is invalid" },
		{ 318, "XIA_WRONG_TYPE - "
		"This routine only operates on detChans that are sets" },
		{ 319, "XIA_UNKNOWN_BOARD - Board type is unknown" },
		{ 320, "XIA_NO_DETCHANS - No detChans are currently defined" },
		{ 321, "XIA_NOT_FOUND - "
			"Unable to locate the Acquisition value requested" },
		{ 322, "XIA_PTR_CHECK - "
			"Pointer is out of synch when it should be valid" },
		{ 323, "XIA_LOOKING_PTRR - "
	"FirmwareSet has a FDD file defined and this only works with PTRRs" },
		{ 324, "XIA_NO_FILENAME - "
			"Requested filename information is set to NULL" },
		{ 325, "XIA_BAD_INDEX - "
			"User specified an alias index that doesn't exist" },
		{ 350, "XIA_FIRM_BOTH - "
"A FirmwareSet may not contain both an FDD and seperate Firmware definitions" },
		{ 351, "XIA_PTR_OVERLAP - "
	"Peaking time ranges in the Firmware definitions may not overlap" },
		{ 352, "XIA_MISSING_FIRM - "
	"Either the FiPPI or DSP file is missing from a Firmware element" },
		{ 353, "XIA_MISSING_POL - "
	"A polarity value is missing from a Detector element" },
		{ 354, "XIA_MISSING_GAIN - "
	"A gain value is missing from a Detector element" },
		{ 355, "XIA_MISSING_INTERFACE - "
	"The interface this channel requires is missing" },
		{ 356, "XIA_MISSING_ADDRESS - "
	"The epp_address information is missing for this channel" },
		{ 357, "XIA_INVALID_NUMCHANS - "
	"The wrong number of channels are assigned to this module" },
		{ 358, "XIA_INCOMPLETE_DEFAULTS - "
	"Some of the required defaults are missing" },
		{ 359, "XIA_BINS_OOR - "
	"There are too many or too few bins for this module type" },
		{ 360, "XIA_MISSING_TYPE - "
	"The type for the current detector is not specified properly" },
		{ 361, "XIA_NO_MMU - "
	"No MMU defined and/or required for this module" },
		{ 401, "XIA_NOMEM - Unable to allocate memory" },
		{ 402, "XIA_XERXES - XerXes returned an error" },
		{ 403, "XIA_MD - MD layer returned an error" },
		{ 404, "XIA_EOF - EOF encountered" },
		{ 405, "XIA_XERXES_NORMAL_RUN_ACTIVE - "
			"XerXes says a normal run is still active" },
		{ 406, "XIA_HARDWARE_RUN_ACTIVE - "
			"The hardware says a control run is still active" },
		{ 501, "XIA_UNKNOWN" },
		{ 502, "XIA_LOG_LEVEL - Log level invalid" },
		{ 503, "XIA_NO_LIST - List size is zero" },
		{ 504, "XIA_NO_ELEM - No data to remove" },
		{ 505, "XIA_DATA_DUP - Data already in table" },
		{ 506, "XIA_REM_ERR - Unable to remove entry from hash table" },
		{ 507, "XIA_FILE_TYPE - Improper file type specified" },
		{ 508, "XIA_END - "
	"There are no more instances of the name specified. Pos set to end" },
		{ 601, "XIA_NOSUPPORT_FIRM - "
	"The specified firmware is not supported by this board type" },
		{ 602, "XIA_UNKNOWN_FIRM - "
			"The specified firmware type is unknown" },
		{ 603, "XIA_NOSUPPORT_VALUE - "
			"The specified acquisition value is not supported" },
		{ 604, "XIA_UNKNOWN_VALUE - "
			"The specified acquisition value is unknown" },
		{ 605, "XIA_PEAKINGTIME_OOR - "
	"The specified peaking time is out-of-range for this product" },
		{ 606, "XIA_NODEFINE_PTRR - "
	"The specified peaking time does not have a PTRR associated with it" },
		{ 607, "XIA_THRESH_OOR - "
			"The specified treshold is out-of-range" },
		{ 608, "XIA_ERROR_CACHE - "
			"The data in the values cache is out-of-sync" },
		{ 609, "XIA_GAIN_OOR - "
			"The specified gain is out-of-range for this produce" },
		{ 610, "XIA_TIMEOUT - Timeout waiting for BUSY" },
		{ 611, "XIA_BAD_SPECIAL - "
	"The specified special run is not supported for this module" },
		{ 612, "XIA_TRACE_OOR - "
	"The specified value of tracewait (in ns) is out-of-range" },
		{ 613, "XIA_DEFAULTS - "
	"The PSL layer encountered an error creating a Defaults element" },
		{ 614, "XIA_BAD_FILTER - Error loading filter info "
	"from either a FDD file or the Firmware configuration" },
		{ 615, "XIA_NO_REMOVE - Specified acquisition value "
	"is required for this product and can't be removed" },
		{ 616, "XIA_NO_GAIN_FOUND - "
	"Handel was unable to converge on a stable gain value" },
		{ 617, "XIA_UNDEFINED_RUN_TYPE - "
	"Handel does not recognize this run type" },
		{ 618, "XIA_INTERNAL_BUFFER_OVERRUN - "
	"Handel attempted to overrun an internal buffer boundry" }, 
		{ 619, "XIA_EVENT_BUFFER_OVERRUN - "
	"Handel attempted to overrun the event buffer boundry" }, 
		{ 620, "XIA_BAD_DATA_LENGTH - "
	"Handel was asked to set a Data length to zero for readout" },
		{ 621, "XIA_NO_LINEAR_FIT - "
	"Handel was unable to perform a linear fit to some data" },
		{ 622, "XIA_MISSING_PTRR - Required PTRR is missing" },
		{ 623, "XIA_PARSE_DSP - Error parsing DSP" },
		{ 624, "XIA_UDXPS", },
		{ 625, "XIA_BIN_WIDTH - Specified bin width is out-of-range" },
		{ 626, "XIA_NO_VGA - "
"An attempt was made to set the gaindac on a board that doesn't have a VGA" },
		{ 627, "XIA_TYPEVAL_OOR - "
	"Specified detector type value is out-of-range" },
		{ 628, "XIA_LOW_LIMIT_OOR - "
	"Specified low MCA limit is out-of-range" },
		{ 629, "XIA_BPB_OOR - bytes_per_bin is out-of-range" },
		{ 630, "XIA_FIP_OOR - Specified FiPPI is out-fo-range" },
		{ 701, "XIA_XUP_VERSION - XUP version is not supported" },
		{ 702, "XIA_CHKSUM - checksum mismatch in the XUP" },
		{ 703, "XIA_BAK_MISSING - Requested BAK file cannot be opened"},
		{ 704, "XIA_SIZE_MISMATCH - "
			"Size read from file is incorrect" },
		{ 1001,"XIA_DEBUG" }
	};

	static const char unrecognized_error_message[]
		= "Unrecognized Handel error code";

	static int num_error_codes = sizeof( error_code_table )
				/ sizeof( error_code_table[0] );

	unsigned int i;

	for ( i = 0; i < num_error_codes; i++ ) {

		if ( handel_status == error_code_table[i].handel_error_code )
			break;		/* Exit the for() loop. */
	}

	if ( i >= num_error_codes ) {
		return &unrecognized_error_message[0];
	}

	return &( error_code_table[i].handel_error_message[0] );
}

#ifndef MX_PROCESS_GET

#define MX_PROCESS_GET	1
#define MX_PROCESS_PUT	2

#endif

static mx_status_type
mxi_xia_handel_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	static const char fname[] = "mxi_xia_handel_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_XIA_HANDEL *xia_handel;
	mx_status_type mx_status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;
	xia_handel = (MX_XIA_HANDEL *) (record->record_type_struct);

	mx_status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( record_field->label_value ) {
		case MXLV_XIA_HANDEL_CONFIG_FILENAME:
			/* Nothing to do since the necessary filename is
			 * already stored in the 'config_filename' field.
			 */

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	case MX_PROCESS_PUT:
		switch( record_field->label_value ) {
		case MXLV_XIA_HANDEL_CONFIG_FILENAME:
			mx_status = mxi_xia_handel_load_config( xia_handel );

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return mx_status;
}

#endif /* HAVE_XIA_HANDEL */
