/*
 * Name:    mx_hrt.c
 *
 * Purpose: This file contains functions that manipulate times shorter
 *          than the scheduling interval for the operating system.
 *
 *          In general, the mx_sleep(), mx_msleep(), and mx_usleep() family
 *          of functions perform their tasks by yielding control to the
 *          operating system and requesting that they be woken up again
 *          by the OS at a later time.  This usually means that you cannot
 *          create delays this way that are shorter than the minimum process
 *          scheduling interval for the operating system (100 Hz for most
 *          versions of Linux).
 *
 *          mx_udelay() instead attempts to implement microsecond delays
 *          that do _not_ explicitly return control to the operating system
 *          by using a busy-wait loop to wait until the requested time has
 *          elapsed.  This allows for delays that are much shorter than the
 *          scheduler would permit.  This is not possible on all operating
 *          systems, so mx_udelay() reverts back to using mx_usleep() in
 *          that case.
 *
 *          mx_high_resolution_time() exports the internal clock used by
 *          mx_udelay() in a 'struct timespec' structure.  The matching
 *          function mx_high_resolution_time_in_seconds() reports back
 *          the same value in seconds using a double variable.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002-2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_clock.h"
#include "mx_hrt.h"

/* The time returned by mx_high_resolution_time_using_clock_ticks() is
 * generally "high resolution" in name only.  But this function serves
 * as a fallback if there is no better method.
 */

static struct timespec
mx_high_resolution_time_using_clock_ticks( void )
{
	MX_CLOCK_TICK clock_tick;
	struct timespec value;
	double time_in_seconds;

	clock_tick = mx_current_clock_tick();

	time_in_seconds = mx_convert_clock_ticks_to_seconds( clock_tick );

	value.tv_sec = (long) time_in_seconds;

	time_in_seconds -= (double) value.tv_sec;

	value.tv_nsec = (long) ( 1000000000.0 * time_in_seconds );

	return value;
}

MX_EXPORT struct timespec
mx_add_high_resolution_times( struct timespec time1,
				struct timespec time2 )
{
	struct timespec result;
	long nanoseconds;

	nanoseconds = time1.tv_nsec + time2.tv_nsec;

	result.tv_nsec = nanoseconds % 1000000000L;

	result.tv_sec = time1.tv_sec + time2.tv_sec + nanoseconds / 1000000000L;

	return result;
}

MX_EXPORT struct timespec
mx_subtract_high_resolution_times( struct timespec time1,
				struct timespec time2 )
{
	struct timespec result;

	if ( time1.tv_nsec < time2.tv_nsec ) {
		time1.tv_sec  -= 1L;
		time1.tv_nsec += 1000000000L;
	}

	result.tv_nsec = time1.tv_nsec - time2.tv_nsec;

	result.tv_sec = time1.tv_sec - time2.tv_sec;

	return result;
}

#if defined(OS_LINUX) && defined(__i386__) && defined(__GNUC__)

/******* x86 Linux *******/

/* The following was inspired by the Linux I/O port programming mini-HOWTO. */

#define MX_CPU_UNKNOWN		0
#define MX_CPU_USE_XCHG		1
#define MX_CPU_USE_RDTSC	2

static int mx_cpu_delay_type = MX_CPU_UNKNOWN;

static double mx_delay_instructions_per_microsecond = 0.0;

static __inline__ unsigned long long int
mx_rdtsc( void )
{
	unsigned long long int x;

	/* The following generates inline an RDTSC instruction
	 * and puts the results in the variable "x".
	 */

	__asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));

	return x;
}

static __inline__ void
mx_xchg_delay( void )
{
	/* The following takes three clock cycles on i386 and i486 machines.
	 * We will be using RDTSC on other x86 CPU types.
	 */

	__asm__ volatile ("xchg %bx,%bx");

	return;
}

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_cpu_delay_type == MX_CPU_UNKNOWN )
		mx_high_resolution_time_init();

	if ( mx_cpu_delay_type == MX_CPU_USE_RDTSC ) {

		/* Warning: The following logic has the fundamental flaw
		 * that the TSC counter will overflow in around 584 years
		 * on a 1 GHz x86 machine, but I expect to be retired long
		 * before that.  (WML)
		 */

		unsigned long long int tsc_start_value, tsc_end_value;
		unsigned long long int tsc_delay_ticks;

		tsc_delay_ticks = (unsigned long long int)
	( mx_delay_instructions_per_microsecond * (double) microseconds );

		tsc_start_value = mx_rdtsc();

		tsc_end_value = tsc_start_value + tsc_delay_ticks;

		/* Sit in a loop until we reach the end of the delay time. */

		while ( mx_rdtsc() < tsc_end_value )
			;				/* Do nothing */

	} else {
		/* This code does not work as well as the RDTSC code since
		 * it does not take into account delays due to the process
		 * being preempted by the kernel.
		 */

		unsigned long long int i, num_times_to_loop;

		num_times_to_loop = (unsigned long long int)
	( mx_delay_instructions_per_microsecond * (double) microseconds );

		for ( i = 0; i < num_times_to_loop; i++ ) {
			mx_xchg_delay();
		}
	}
	return;
}

MX_EXPORT struct timespec
mx_high_resolution_time( void )
{
	if ( mx_cpu_delay_type == MX_CPU_UNKNOWN )
		mx_high_resolution_time_init();

	if ( mx_cpu_delay_type != MX_CPU_USE_RDTSC ) {
		return mx_high_resolution_time_using_clock_ticks();
	} else {
		struct timespec value;
		unsigned long long tsc_value;
		double time_in_microseconds;

		tsc_value = mx_rdtsc();

		time_in_microseconds = mx_divide_safely( (double) tsc_value,
					mx_delay_instructions_per_microsecond );

		value.tv_sec = (unsigned long)
				( 0.000001 * time_in_microseconds );

		time_in_microseconds -= 1000000.0 * (double) value.tv_sec;

		value.tv_nsec = (unsigned long)
				( 1000.0 * time_in_microseconds );

		return value;
	}
}

MX_EXPORT void
mx_high_resolution_time_init( void )
{
	static const char fname[] = "mx_high_resolution_time_init()";

	char buffer[500];
	char *ptr, *ptr2;
	int cpu_family, have_tsc, num_items;
	double cpu_mhz;

	/* Reading /proc/cpuinfo is the only vaguely portable way I know of
	 * for getting the necessary information.
	 */

	FILE *cpuinfo;

	cpuinfo = fopen( "/proc/cpuinfo", "r" );

	if ( cpuinfo == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"/proc/cpuinfo could not be found or is unreadable.  "
		"This may indicate that there is something seriously "
		"wrong with your Linux system." );

		return;
	}

	cpu_family = 0;
	have_tsc = FALSE;
	cpu_mhz = 0.0;

	fgets( buffer, sizeof buffer, cpuinfo );

	while ( !feof(cpuinfo) && !ferror(cpuinfo) ) {

		if ( strncmp( buffer, "cpu family", 10 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				num_items = sscanf( ptr, "%d", &cpu_family );

				if ( num_items != 1 ) {
				    (void) mx_error( MXE_UNPARSEABLE_STRING,
					fname,
				"Could not find the cpu family number in a "
				"line read from /proc/cpuinfo.  line = '%s'",
					buffer );
				}
			}
		} else
		if ( strncmp( buffer, "cpu MHz", 7 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				num_items = sscanf( ptr, "%lg", &cpu_mhz );

				if ( num_items != 1 ) {
				    (void) mx_error( MXE_UNPARSEABLE_STRING,
					fname,
				"Could not find the cpu MHz in a "
				"line read from /proc/cpuinfo.  line = '%s'",
					buffer );
				}
			}
		} else
		if ( strncmp( buffer, "flags", 5 ) == 0 ) {

			ptr = strchr( buffer, ':' );

			if ( ptr == NULL ) {
				(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
				"Could not find a ':' character in a line "
				"read from /proc/cpuinfo.  line = '%s'",
					buffer );
			} else {
				ptr++;

				ptr2 = strstr( ptr, "tsc" );

				if ( ptr2 != NULL ) {
					have_tsc = TRUE;
				}
			}
		}

		fgets( buffer, sizeof buffer, cpuinfo );
	}

	fclose( cpuinfo );

	if ( have_tsc ) {
		mx_cpu_delay_type = MX_CPU_USE_RDTSC;
	} else {
		mx_cpu_delay_type = MX_CPU_USE_XCHG;
	}

	if ( (cpu_family == 3) || (cpu_family == 4) ) {
		mx_delay_instructions_per_microsecond = cpu_mhz / 3.0;
	} else {
		mx_delay_instructions_per_microsecond = cpu_mhz;
	}

	return;

}

#elif defined(OS_WIN32)

/******* Microsoft Win32 *******/

#include <windows.h>

static int mx_high_resolution_time_init_invoked = FALSE;

static int mx_performance_counters_are_available = FALSE;

static double mx_performance_counter_ticks_per_microsecond = 0.0;

static void
mx_hrt_error_message( const char *function_name )
{
	static const char fname[] = "mx_hrt_error_message()";

	DWORD last_error_code;
	TCHAR message_buffer[MXU_ERROR_MESSAGE_LENGTH - 120];

	last_error_code = GetLastError();

	mx_win32_error_message( last_error_code,
		message_buffer, sizeof(message_buffer) );

	(void) mx_error( MXE_FUNCTION_FAILED, fname,
	"Error invoking %s.  Win32 error code = %ld, error message = '%s'",
		function_name, last_error_code, message_buffer );

	return;
}

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_high_resolution_time_init_invoked == FALSE )
		mx_high_resolution_time_init();

	if ( mx_performance_counters_are_available == FALSE ) {
		mx_usleep( microseconds );
	} else {
		LARGE_INTEGER counter_value;
		LONGLONG counter_start_value, counter_end_value;
		LONGLONG counter_current_value, counter_delay_ticks;
		BOOL status;

		counter_delay_ticks = (LONGLONG)
	( mx_performance_counter_ticks_per_microsecond * (double)microseconds );

		status = QueryPerformanceCounter( &counter_value );

		if ( status == 0 ) {
			mx_hrt_error_message( "QueryPerformanceCounter()" );
			return;
		}

		counter_start_value = counter_value.QuadPart;

		counter_end_value = counter_start_value + counter_delay_ticks;

		/* Sit in a loop until we reach the end of the delay time. */

		counter_current_value = counter_start_value;

		while ( counter_current_value < counter_end_value ) {

			status = QueryPerformanceCounter( &counter_value );

			if ( status == 0 ) {
				mx_hrt_error_message(
					"QueryPerformanceCounter()" );
				return;
			}

			counter_current_value = counter_value.QuadPart;
		}
	}
	return;
}

MX_EXPORT struct timespec
mx_high_resolution_time( void )
{
	if ( mx_high_resolution_time_init_invoked == FALSE )
		mx_high_resolution_time_init();

	if ( mx_performance_counters_are_available == FALSE ) {

		return mx_high_resolution_time_using_clock_ticks();

	} else {
		struct timespec value;
		double time_in_microseconds;
		LARGE_INTEGER counter_value;
		BOOL status;

		status = QueryPerformanceCounter( &counter_value );

		if ( status == 0 ) {
			value.tv_sec = 0;
			value.tv_nsec = 0;

			mx_hrt_error_message( "QueryPerformanceCounter()" );

			return value;
		}

		time_in_microseconds = mx_divide_safely(
				(double) counter_value.QuadPart,
				mx_performance_counter_ticks_per_microsecond );

		value.tv_sec = (unsigned long)
				( 0.000001 * time_in_microseconds );

		time_in_microseconds -= 1000000.0 * (double) value.tv_sec;

		value.tv_nsec = (unsigned long)
				( 1000.0 * time_in_microseconds );

		return value;
	}
}

MX_EXPORT void
mx_high_resolution_time_init( void )
{
	LARGE_INTEGER performance_frequency;
	BOOL status;

	mx_high_resolution_time_init_invoked = TRUE;

	status = QueryPerformanceFrequency( &performance_frequency );

	if ( status == 0 ) {
		mx_hrt_error_message( "QueryPerformanceFrequency()" );

		mx_performance_counters_are_available = FALSE;

		mx_warning(
			"Win32 performance counters are not available, "
			"so we will revert to using mx_usleep() instead." );
	} else {
		mx_performance_counters_are_available = TRUE;

		mx_performance_counter_ticks_per_microsecond
			= 1.0e-6 * (double) performance_frequency.QuadPart;
	}
	return;
}

#else

/******* None of the above ********/

static int mx_high_resolution_time_init_invoked = FALSE;

MX_EXPORT void
mx_udelay( unsigned long microseconds )
{
	if ( mx_high_resolution_time_init_invoked == FALSE )
		mx_high_resolution_time_init();

	mx_usleep( microseconds );
}

MX_EXPORT struct timespec
mx_high_resolution_time( void )
{
	return mx_high_resolution_time_using_clock_ticks();
}

MX_EXPORT void
mx_high_resolution_time_init( void )
{
	mx_high_resolution_time_init_invoked = TRUE;

	mx_warning( "High resolution time measurement is not yet implemented "
		"on this operating system." );
	return;
}

#endif
