/*
 * Name:    mx_scan_linear.c
 *
 * Purpose: Driver for MX_LINEAR_SCAN step scan records.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001, 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_alloc.h"
#include "mx_scan.h"
#include "mx_scan_linear.h"

MX_RECORD_FUNCTION_LIST mxs_linear_scan_record_function_list = {
	mxs_linear_scan_initialize_type,
	mxs_linear_scan_create_record_structures,
	mxs_linear_scan_finish_record_initialization,
	mxs_linear_scan_delete_record,
	mxs_linear_scan_print_scan_structure
};

MX_SCAN_FUNCTION_LIST mxs_linear_scan_scan_function_list = {
	mxs_linear_scan_prepare_for_scan_start,
	mxs_linear_scan_execute_scan_body,
	mxs_linear_scan_cleanup_after_scan_end
};

/* Prototype an internal use only function. */

static mx_status_type
mxs_linear_scan_execute_scan_level( MX_SCAN *scan,
		MX_LINEAR_SCAN *linear_scan,
		long dimension_level,
		mx_status_type (*compute_motor_positions_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *),
		mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_MOTOR_MOVE_REPORT_FUNCTION, int)
		);

/*=========*/

#define NUM_LINEAR_SCAN_FIELDS	4

MX_EXPORT mx_status_type
mxs_linear_scan_initialize_type( long record_type )
{
	const char fname[] = "mxs_linear_scan_initialize_type()";

	const char field_name[NUM_LINEAR_SCAN_FIELDS][MXU_FIELD_NAME_LENGTH+1]
	    = {
		"start_position",
		"step_size",
		"num_measurements",
		"step_number"};

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	int i;
	long num_record_fields;
	long num_independent_variables_varargs_cookie;
	long num_motors_varargs_cookie;
	long num_input_devices_varargs_cookie;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	/**** Fix up the record fields common to all scan types. ****/

	num_record_fields = *(driver->num_record_fields);

	status = mx_scan_fixup_varargs_record_field_defaults(
			record_field_defaults, num_record_fields,
			&num_independent_variables_varargs_cookie,
			&num_motors_varargs_cookie,
			&num_input_devices_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	/**** Fix up the record fields specific to MX_LINEAR_SCAN records ****/

	for ( i = 0; i < NUM_LINEAR_SCAN_FIELDS; i++ ) {
		status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			field_name[i], &field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field->dimension[0] = num_independent_variables_varargs_cookie;
	}

	return MX_SUCCESSFUL_RESULT;
}

#undef NUM_LINEAR_SCAN_FIELDS		/* Just for safety's sake. */

MX_EXPORT mx_status_type
mxs_linear_scan_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_create_record_structures()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	MX_DRIVER *driver;
	mx_status_type (*fptr)( MX_RECORD *, MX_SCAN *, MX_LINEAR_SCAN * );
	mx_status_type status;

	/* Allocate memory for the necessary structures. */

	scan = (MX_SCAN *) malloc( sizeof(MX_SCAN) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAN structure." );
	}

	linear_scan = (MX_LINEAR_SCAN *) malloc( sizeof(MX_LINEAR_SCAN) );

	if ( linear_scan == (MX_LINEAR_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_LINEAR_SCAN structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_superclass_struct = scan;
	record->record_class_struct = linear_scan;
	record->record_type_struct = NULL;

	scan->record = record;

	scan->measurement.scan = scan;
	scan->datafile.scan = scan;
	scan->plot.scan = scan;

	record->superclass_specific_function_list
				= &mxs_linear_scan_scan_function_list;

	/* Find the MX_LINEAR_SCAN_FUNCTION_LIST pointer. */

	driver = mx_get_driver_by_type( mx_type_list, record->mx_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DRIVER for record type %ld is NULL.",
			record->mx_type );
	}

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
			(driver->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST for record type %ld is NULL.",
			record->mx_type );
	}

	record->class_specific_function_list = flist;

	/* Call the type specific 'create_record_structures' function. */

	fptr = flist->create_record_structures;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Linear scan type specific 'create_record_structures' pointer is NULL.");
	}

	status = (*fptr)( record, scan, linear_scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_finish_record_initialization()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_RECORD * );
	long dimension[2];
	size_t element_size[2];
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	status = mx_scan_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	scan = (MX_SCAN *) record->record_superclass_struct;

	/* If needed allocate memory for the alternate motor position arrays. */

	if ( scan->datafile.num_x_motors > 0 ) {
		dimension[0] = scan->datafile.num_x_motors;
		dimension[1] = 1;

		element_size[0] = sizeof(double);
		element_size[1] = sizeof(double *);

		scan->datafile.x_position_array = (double **)
			mx_allocate_array( 2, dimension, element_size );

		if ( scan->datafile.x_position_array == (double **) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld by %ld array of "
		"alternate position values for scan '%s'.",
				dimension[0], dimension[1], scan->record->name);
		}
	}

	if ( scan->plot.num_x_motors > 0 ) {
		dimension[0] = scan->plot.num_x_motors;
		dimension[1] = 1;

		element_size[0] = sizeof(double);
		element_size[1] = sizeof(double *);

		scan->plot.x_position_array = (double **)
			mx_allocate_array( 2, dimension, element_size );

		if ( scan->plot.x_position_array == (double **) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld by %ld array of "
		"alternate position values for scan '%s'.",
				dimension[0], dimension[1], scan->record->name);
		}
	}

	/* Call the record type specific 'finish_record_initialization'
	 * function.
	 */

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST pointer is NULL.");
	}

	fptr = flist->finish_record_initialization;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Linear scan type specific 'finish_record_initialization' pointer is NULL.");
	}

	status = (*fptr)( record );

	return status;
}

MX_EXPORT mx_status_type
mxs_linear_scan_delete_record( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_delete_record()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_RECORD * );
	long dimension[2];
	size_t element_size[2];
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	scan = (MX_SCAN *)(record->record_superclass_struct);

	if ( scan != NULL ) {

		linear_scan = (MX_LINEAR_SCAN *)
					(scan->record->record_class_struct);

		if ( linear_scan != NULL ) {
			flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

			if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_LINEAR_SCAN_FUNCTION_LIST pointer is NULL.");
			}

			fptr = flist->delete_record;

			/* If fptr is NULL, this record type has no type
			 * specific structures to delete.  Otherwise,
			 * we invoke fptr to delete the type specific
			 * structures.
			 */

			if ( fptr != NULL ) {
				status = (*fptr)( record );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			record->record_type_struct = NULL;
			linear_scan->record_type_struct = NULL;

			free( linear_scan );

			record->record_class_struct = NULL;
		}

		if ( scan->datafile.x_motor_array != NULL ) {
			free( scan->datafile.x_motor_array );
		}

		if ( scan->datafile.x_position_array != NULL ) {
			dimension[0] = scan->datafile.num_x_motors;
			dimension[1] = 1;

			element_size[0] = sizeof(double);
			element_size[1] = sizeof(double *);

			(void) mx_free_array( scan->datafile.x_position_array,
					      2, dimension, element_size );
		}

		if ( scan->plot.x_motor_array != NULL ) {
			free( scan->plot.x_motor_array );
		}

		if ( scan->plot.x_position_array != NULL ) {
			dimension[0] = scan->plot.num_x_motors;
			dimension[1] = 1;

			element_size[0] = sizeof(double);
			element_size[1] = sizeof(double *);

			(void) mx_free_array( scan->plot.x_position_array,
					      2, dimension, element_size );
		}

		free( scan );

		record->record_superclass_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_print_scan_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_print_scan_structure()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN *linear_scan;
	MX_MOTOR *motor;
	long i, j;
	double end_position;
	mx_status_type status;

	if ( record != NULL ) {
		fprintf( file, "SCAN parameters for linear scan '%s':\n",
				record->name );
	}

	status = mx_scan_print_scan_structure( file, record );

	if ( status.code != MXE_SUCCESS )
		return status;

	scan = (MX_SCAN *) record->record_superclass_struct;

	linear_scan = (MX_LINEAR_SCAN *) record->record_class_struct;

	if ( linear_scan == (MX_LINEAR_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN pointer for scan '%s' is NULL.",
			record->name );
	}

	fprintf( file, "  Scan range:\n" );

	j = 0;

	if ( scan->num_motors == 0 ) {
		if ( scan->num_independent_variables > 0 ) {
			fprintf( file, "    %ld measurements\n",
				linear_scan->num_measurements[j] );
		}
	} else {
		for ( i = 0; i < scan->num_motors; i++ ) {
			if ( scan->motor_is_independent_variable[i] == TRUE ) {
				motor = (MX_MOTOR *)
			    scan->motor_record_array[i]->record_class_struct;
	
				end_position = linear_scan->start_position[j]
					+ linear_scan->step_size[j] * (double)
				    ( linear_scan->num_measurements[j] - 1 );
	
				fprintf( file,
		"    %s: %g %s to %g %s, %g %s step size, %ld measurements\n",
					scan->motor_record_array[i]->name,
					linear_scan->start_position[j],
					motor->units,
					end_position,
					motor->units,
					linear_scan->step_size[j],
					motor->units,
					linear_scan->num_measurements[j] );
	
				j++;
			}
		}
	}
	fprintf( file, "\n" );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_linear_scan_prepare_for_scan_start()";

	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr) (MX_SCAN *);
	int i;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked", fname));

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(scan->record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST pointer for scan '%s' is NULL.",
			scan->record->name );
	}
	
	/* Some motor drivers use cached values of variables when 
	 * scan 'fast mode' is turned on rather than querying a
	 * remote server for the values on each step.  For example,
	 * the 'energy' pseudomotor does not reread the value of
	 * 'd_spacing' on each step when 'fast mode' is turned on.
	 *
	 * We can make sure that the cached values are up to date
	 * by reading the position of each motor used by the scan
	 * before 'fast mode' is turned on.
	 */

	for ( i = 0; i < scan->num_motors; i++ ) {
		status = mx_motor_get_position( scan->motor_record_array[i],
						NULL );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Perform standard scan startup operations.
	 *
	 * Among other things, this turns on 'fast mode'.
	 */

	status = mx_standard_prepare_for_scan_start( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Clear all the input devices that need it. */

	status = mx_clear_scan_input_devices( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Perform scan type specific initialization. */

	fptr = flist->prepare_for_scan_start;

	/* It is not an error for flist->prepare_for_scan_start to be NULL,
	 * this just means that this particular scan type does not need
	 * to do anything special when a scan is starting.
	 */

	if ( fptr != NULL ) {
		status = (*fptr) (scan);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_execute_scan_body( MX_SCAN *scan )
{
	const char fname[] = "mxs_linear_scan_execute_scan_body()";

	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*compute_motor_positions_fptr)
					( MX_SCAN *, MX_LINEAR_SCAN * );
	mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_MOTOR_MOVE_REPORT_FUNCTION, int);
	mx_status_type status;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_SCAN pointer passed was NULL." );
	}

	linear_scan = (MX_LINEAR_SCAN *)(scan->record->record_class_struct);

	if ( linear_scan == (MX_LINEAR_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
			(scan->record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_LINEAR_SCAN_FUNCTION_LIST ptr for scan record '%s' is NULL.",
			scan->record->name );
	}

	compute_motor_positions_fptr = flist->compute_motor_positions;

	if ( compute_motor_positions_fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"'compute_motor_position' function pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	/* It is not an error for flist->motor_record_array_move_special
	 * to be NULL.  If this function ptr is NULL, this just means that
	 * mx_motor_record_array_move_absolute will be called directly
	 * rather than being preprocessed by motor_record_array_move_special.
	 */

	move_special_fptr = flist->motor_record_array_move_special;

	status = mxs_linear_scan_execute_scan_level( scan, linear_scan,
			scan->num_independent_variables - 1,
			compute_motor_positions_fptr,
			move_special_fptr );

	return status;
}

MX_EXPORT mx_status_type
mxs_linear_scan_cleanup_after_scan_end( MX_SCAN *scan )
{
	return mx_standard_cleanup_after_scan_end( scan );
}

static mx_status_type
mxs_linear_scan_execute_scan_level( MX_SCAN *scan,
		MX_LINEAR_SCAN *linear_scan,
		long dimension_level,
		mx_status_type (*compute_motor_positions_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *),
		mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_MOTOR_MOVE_REPORT_FUNCTION, int)
		)
{
	const char fname[] = "mxs_linear_scan_execute_scan_level()";

	MX_RECORD *x_motor_record;
	double x_motor_position;
	long i, j, k, array_index, num_dimension_steps, new_dimension_level;
	int get_position;
	mx_status_type status;

	array_index = scan->num_independent_variables - dimension_level - 1;

	num_dimension_steps = linear_scan->num_measurements[ array_index ];

	new_dimension_level = dimension_level - 1;

	MX_DEBUG( 2,
("%s: dimension_level = %ld, array_index = %ld, num_dimension_steps = %ld",
		fname, dimension_level, array_index, num_dimension_steps ));

	if ( dimension_level > 0 ) {
		/** More levels in the array to descend through. **/

		MX_DEBUG( 2,("%s: -> Go to next dimension level = %ld",
					fname, new_dimension_level));

		for ( i = 0; i < num_dimension_steps; i++ ) {

			linear_scan->step_number[ array_index ] = i;

			status = mxs_linear_scan_execute_scan_level(scan,
					linear_scan,
					new_dimension_level,
					compute_motor_positions_fptr,
					move_special_fptr);

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	} else {
		/** At bottom level of independent variables. **/

		MX_DEBUG( 2,
		("%s: At bottom level of independent variables.", fname ));

		if ( scan->num_motors > 0 ) {
			mx_info("Moving to start position for scan section.");
		}

		for ( i = 0; i < num_dimension_steps; i++ ) {

			/* Compute the next set of motor positions. */

			linear_scan->step_number[ array_index ] = i;

			status = (*compute_motor_positions_fptr)
						( scan, linear_scan );

			if ( status.code != MXE_SUCCESS )
				return status;

			do {		/** Pause/abort retry loop. **/

				/* Move to the computed motor positions. */

				/* If move_special_fptr is NULL, we just invoke
				 * mx_motor_array_move_absolute directly.
				 * Otherwise, the computed move must be
				 * preprocessed by move_special_fptr before
				 * the motors are moved.  See how slit scans
				 * are implemented in 'libmx/s_slit.c' for
				 * an example of this.
				 */

				if ( move_special_fptr == NULL ) {
					if ( scan->num_motors > 0 ) {
						status = 
				    mx_motor_array_move_absolute_with_report(
						scan->num_motors,
						scan->motor_record_array,
						scan->motor_position,
						NULL,
						MXF_MTR_SCAN_IN_PROGRESS );
					}
				} else {
					status = (*move_special_fptr)
						( scan,
						linear_scan,
						scan->num_motors,
						scan->motor_record_array,
						scan->motor_position,
						NULL,
						MXF_MTR_SCAN_IN_PROGRESS );
				}

				if ( ( status.code != MXE_SUCCESS )
				  && ( status.code != MXE_PAUSE_REQUESTED ) )
				{
					return status;
				}

				/* Perform the measurement. */

				if ( status.code != MXE_PAUSE_REQUESTED ) {

				    status = mx_scan_handle_data_measurement(
									scan );

					if ( ( status.code != MXE_SUCCESS )
					  && ( status.code !=
							MXE_PAUSE_REQUESTED ) )
					{
						return status;
					}
				}

				/* If a pause has not been requested by the
				 * user, exit the pause/abort retry loop.
				 */

				if ( status.code == MXE_SUCCESS ) {
					break;   /* Exit the do..while loop */
				}

				/* If we get here, the most recent status
				 * code was MXE_PAUSE_REQUESTED, so invoke
				 * a function to handle the pause request.
				 */

				status = mx_scan_handle_pause_request( scan );

				if ( status.code != MXE_SUCCESS )
					return status;

				mx_info( "Retrying the last scan step." );

				status = mx_wait_for_motor_array_stop(
						scan->num_motors,
						scan->motor_record_array,
						MXF_MTR_SCAN_IN_PROGRESS );

				if ( status.code != MXE_SUCCESS )
					return status;

			} while (1);	/** End of pause/abort retry loop. **/

			/* If alternate X axis motors have been specified,
			 * get and save their current positions for use
			 * by the datafile handling code.
			 *
			 * Note that x_position_array is an N by 1 array,
			 * where N is the number of motors, since we only
			 * handle one measurement at a time in step scans.
			 */

			for ( j = 0; j < scan->datafile.num_x_motors; j++ ) {

			    x_motor_record = scan->datafile.x_motor_array[j];

			    status = mx_motor_get_position( x_motor_record,
							&x_motor_position );

			    if ( status.code != MXE_SUCCESS )
				    return status;

			    scan->datafile.x_position_array[j][0]
				    = x_motor_position;
			}

			/* Do the same thing for the plot data.
			 *
			 * The list of alternate X motors for the plot will
			 * not necessarily be the same as the list for the
			 * datafile, but we do not want to unnecessarily
			 * read a motor position twice.
			 */

			for ( j = 0; j < scan->plot.num_x_motors; j++ ) {

			    x_motor_record = scan->plot.x_motor_array[j];

			    /* See if this motor's position was already
			     * read by the datafile loop above.
			     */

			    get_position = TRUE;

			    for ( k = 0; k < scan->datafile.num_x_motors; k++ )
			    {
				if ( x_motor_record ==
					    scan->datafile.x_motor_array[j] )
				{
				    scan->plot.x_position_array[j][0] =
					  scan->datafile.x_position_array[k][0];

				    get_position = FALSE;
				    break;	/* Exit the inner for() loop. */
				}
			    }

			    /* If not, then we must read the position now. */

			    if ( get_position ) {
				status = mx_motor_get_position( x_motor_record,
							&x_motor_position );

				if ( status.code != MXE_SUCCESS )
					return status;

				scan->plot.x_position_array[j][0]
					= x_motor_position;
			    }
			}

			/**** Write the result out to the datafile. ****/

			status = mx_add_measurement_to_datafile(
							&(scan->datafile) );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Report the progress of the scan back to the user. */

			status = mx_scan_display_scan_progress( scan );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Finally, if enabled, add the measurement to the
			 * running scan plot.
			 */

			if ( mx_plotting_is_enabled( scan->record ) ) {

				if ( i == 0 ) {
					/* Tell the plotting package that
					 * this is a new section of the scan.
					 */

					status = mx_plot_start_plot_section(
							&(scan->plot) );

					if ( status.code != MXE_SUCCESS )
						return status;
				}

				status = mx_add_measurement_to_plot_buffer(
							&(scan->plot) );

				if ( status.code != MXE_SUCCESS )
					return status;

				status = mx_display_plot( &(scan->plot) );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			status = mx_scan_increment_measurement_number( scan );

			if ( status.code != MXE_SUCCESS )
				return status;
		}

		(scan->plot.section_number)++;
	}
	return MX_SUCCESSFUL_RESULT;
}

