/*
 * Name:    mx_socket.c
 *
 * Purpose: BSD sockets and Winsock interface functions.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_osdef.h"

#if USE_TCPIP

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

#ifdef OS_SOLARIS
# include <sys/filio.h>
#endif

#include "mx_unistd.h"

#include "mx_util.h"
#include "mx_types.h"
#include "mx_record.h"
#include "mx_ascii.h"
#include "mx_socket.h"
#include "mx_net.h"

static int mx_sockets_are_initialized = FALSE;

#if defined( OS_WIN32 )
   static mx_status_type mx_winsock_initialize( void );
   static char * mx_winsock_strerror( int winsock_errno );

#  define mx_socket_initialize	mx_winsock_initialize

#elif defined( OS_DJGPP )
   static mx_status_type mx_watt32_initialize( void );

#  define mx_socket_initialize  mx_watt32_initialize

#else
#  define mx_socket_initialize	mx_successful_result	/* Just a no-op. */

#endif

#ifndef INADDR_NONE
#define INADDR_NONE	0xffffffff
#endif

MX_EXPORT mx_status_type
mx_socket_open_as_client( MX_SOCKET *client_socket,
		char *hostname, int port_number, int open_flags )
{
	static char fname[] = "mx_socket_open_as_client()";

	struct sockaddr_in destination_address;
	unsigned long inet_address;

	int saved_errno, status;
	char *error_string;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( client_socket == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"client_socket pointer passed is NULL." );
	}
	if ( hostname == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Null hostname string pointer given as an argument." );
	}
	if ( port_number <= 0 || port_number > 65535 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal port number %d given as an argument.", port_number );
	}

	if ( mx_sockets_are_initialized == FALSE ) {

		mx_status = mx_socket_initialize();

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_sockets_are_initialized = TRUE;
	}

	mx_status = mx_socket_get_inet_address( hostname, &inet_address );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	destination_address.sin_family = AF_INET;
	destination_address.sin_addr.s_addr = inet_address;
	destination_address.sin_port = htons( (unsigned short) port_number );

	/* Create a socket. */

	*client_socket = socket( AF_INET, SOCK_STREAM, 0 );

	saved_errno = mx_socket_check_error_status(
			client_socket, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Attempt to create a TCP socket failed.  "
			"Errno = %d.  Error string = '%s'.",
			saved_errno, error_string );
	}

	/* Open the socket. */

	MX_DEBUG( 2,("%s: About to invoke connect().",fname));

	status = connect( *client_socket,
			(struct sockaddr *) &destination_address,
			sizeof( destination_address ) );

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"connect() to host '%s', port %d failed.  Errno = %d.  Error string = '%s'.",
			hostname, port_number,
			saved_errno, error_string );
	}

	MX_DEBUG( 2,("Leaving %s.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_socket_open_as_server( MX_SOCKET *server_socket,
				int port_number, int open_flags )
{
	static char fname[] = "mx_socket_open_as_server()";

	struct sockaddr_in server_address;
	int saved_errno, status, reuseaddr;
	int non_blocking;
	char *error_string;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s: port_number = %d, open_flags = %d",
			fname, port_number, open_flags ));

	if ( server_socket == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_socket pointer passed is NULL." );
	}
	if ( port_number <= 0 || port_number > 65535 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal port number %d given as an argument.", port_number );
	}

	if ( mx_sockets_are_initialized == FALSE ) {

		mx_status = mx_socket_initialize();

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_sockets_are_initialized = TRUE;
	}

	/* Create a socket. */

	*server_socket = socket( AF_INET, SOCK_STREAM, 0 );

	saved_errno = mx_socket_check_error_status(
			server_socket, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Attempt to create a TCP socket failed.  "
			"Errno = %d.  Error string = '%s'.",
			saved_errno, error_string );
	}

	/* Allow address reuse. */

	reuseaddr = 1;

	status = setsockopt( *server_socket, SOL_SOCKET,
		SO_REUSEADDR, (char *) &reuseaddr, sizeof(reuseaddr) );

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Attempt to set SO_REUSEADDR option for server socket for port %d failed.  "
			"Errno = %d.  Error string = '%s'.",
			port_number, saved_errno, error_string );
	}

	/* Bind to the local port so that clients can connect to us. */

	memset((char *) &server_address, 0, sizeof( server_address ));

	server_address.sin_family      = AF_INET;
	server_address.sin_addr.s_addr = htonl( INADDR_ANY );
	server_address.sin_port        = htons( (unsigned short) port_number );

	status = bind( *server_socket, (struct sockaddr *) &server_address,
				sizeof( server_address ) );

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Attempt to bind server socket to port %d failed.  "
			"Errno = %d.  Error string = '%s'.",
			port_number, saved_errno, error_string );
	}

	/* Indicate our willingness to accept connections. */

	status = listen( *server_socket, 5 );    /* 5 is traditional. */

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Attempt to listen for connections to port %d failed.  "
			"Errno = %d.  Error string = '%s'.",
			port_number, saved_errno, error_string );
	}

	/* Set the socket to non-blocking mode. */

	non_blocking = TRUE;

#if defined( OS_WIN32 )
	status = ioctlsocket( *server_socket, FIONBIO, &non_blocking );
#elif defined( OS_DJGPP )
	status = ioctlsocket( *server_socket, FIONBIO, (char *) &non_blocking );
#elif defined( OS_VXWORKS )
	status = ioctl( *server_socket, FIONBIO, (int) &non_blocking );
#else
	status = ioctl( *server_socket, FIONBIO, &non_blocking );
#endif

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Error while trying to set socket to non-blocking mode.  "
		"Errno = %d.  Error string = '%s'.",
		saved_errno, error_string );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_socket_close( MX_SOCKET tcp_socket )
{
	static char fname[] = "mx_socket_close()";

	int status, non_blocking, saved_errno;
	char c, *error_string;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* Set the socket to non-blocking mode. */

	non_blocking = TRUE;

#if defined( OS_WIN32 )
	status = ioctlsocket( tcp_socket, FIONBIO, &non_blocking );
#elif defined( OS_DJGPP )
	status = ioctlsocket( tcp_socket, FIONBIO, (char *) &non_blocking );
#elif defined( OS_VXWORKS )
	status = ioctl( tcp_socket, FIONBIO, (int) &non_blocking );
#else
	status = ioctl( tcp_socket, FIONBIO, &non_blocking );
#endif

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Error while trying to set socket to non-blocking mode.  "
		"Errno = %d.  Error string = '%s'.",
		saved_errno, error_string );
	}

	/* "Half-close" the socket by disallowing any further writes
	 * from our end.
	 */

	status = shutdown( tcp_socket, 1 );

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	if ( saved_errno != 0 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Error while executing shutdown( tcp_socket, 1 ).  "
		"Errno = %d.  Error string = '%s'.",
		saved_errno, error_string );
	}

	/* Read from the socket until there is no more data. */

	for(;;) {
		status = recv( tcp_socket, &c, 1, 0 );

		if ( status == 0 ) {
			/* The connection is closed, so we're done. */
			break;		/* exit the for() loop. */

		} else if ( status == MX_SOCKET_ERROR ) {
			saved_errno = mx_socket_get_last_error();
#ifdef OS_WIN32
			if ( saved_errno == WSAEWOULDBLOCK ) {
#else
			if ( saved_errno == EWOULDBLOCK ) {
#endif
				/* Nothing more to read, so we're done. */
				break;	/* exit the for() loop. */
			} else {
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Error during socket close procedure while executing recv().  "
	"Errno = %d.  Error string = '%s'.", saved_errno,
					mx_socket_strerror( saved_errno ) );
			}
		}
	}

	/* Now finish shutting down the connection. */

	status = shutdown( tcp_socket, 2 );

	saved_errno = mx_socket_check_error_status(
			&status, MXF_SOCKCHK_INVALID, &error_string );

	switch ( saved_errno ) {
	case 0:

#ifdef OS_WIN32
	case WSAENOTCONN:		/* The server didn't wait for us. */
#else
	case ENOTCONN:			/* The server didn't wait for us. */
	case EINVAL:
#endif
		/* Nothing to do. */

		break;
	default:
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Error while executing shutdown( tcp_socket, 2 ).  "
			"Errno = %d.  Error string = '%s'.",
			saved_errno, error_string );
		break;
	}

#if defined( OS_WIN32 ) || defined( OS_DJGPP )
	closesocket( tcp_socket );

#elif (defined( OS_VMS ) && defined( HAVE_WINTCP ))
	vms_close( tcp_socket );

#else
	close( tcp_socket );
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT int
mx_socket_is_open( MX_SOCKET tcp_socket )
{
#ifdef OS_WIN32
	if ( tcp_socket == INVALID_SOCKET ) {
		return FALSE;
	} else {
		return TRUE;
	}
#else
	if ( tcp_socket < 0 ) {
		return FALSE;
	} else {
		return TRUE;
	}
#endif
}

MX_EXPORT void
mx_socket_mark_as_closed( MX_SOCKET *tcp_socket )
{
	*tcp_socket = MX_INVALID_SOCKET;
	return;
}

MX_EXPORT char *
mx_socket_strerror( int saved_errno )
{
#ifdef OS_WIN32
	return mx_winsock_strerror( saved_errno );
#else
	return strerror( saved_errno );
#endif
}


MX_EXPORT int
mx_socket_check_error_status( void *value_to_check,
				int type_of_check, char **error_string )
{
	static char fname[] = "mx_socket_check_error_status()";

	int saved_errno, int_value;

#if 0
	MX_DEBUG( 2,("%s invoked.", fname));
#endif

	switch( type_of_check ) {
	case MXF_SOCKCHK_INVALID:
		int_value = *( (int *) value_to_check );

		if ( int_value != MX_INVALID_SOCKET ) {
			*error_string = NULL;
			return 0;
		}
		break;
	case MXF_SOCKCHK_ZERO:
		int_value = *( (int *) value_to_check );

		if ( int_value != 0 ) {
			*error_string = NULL;
			return 0;
		}
		break;
	case MXF_SOCKCHK_NULL:
		if ( value_to_check != NULL ) {
			*error_string = NULL;
			return 0;
		}
		break;
	default:
		mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown type_of_check = %d.", type_of_check );
		*error_string = NULL;
		return -1;
		break;
	}

	saved_errno = mx_socket_get_last_error();

	*error_string = mx_socket_strerror( saved_errno );

	return saved_errno;
}

MX_EXPORT mx_status_type
mx_socket_get_inet_address( char *hostname, unsigned long *inet_address )
{
	static const char fname[] = "mx_socket_get_inet_address()";

	MX_DEBUG( 2,("%s invoked for hostname '%s'.", fname, hostname));

	/* First test to see if hostname is a dotted decimal number.  Note
	 * that 255.255.255.255 will fail this test.
	 */

	*inet_address = inet_addr( hostname );

	if ( *inet_address == INADDR_NONE ) {

		struct hostent *host_entry;
		struct in_addr *in_addr_struct;
		int error_value;
		void *void_ptr;

		/* The hostname is not dotted decimal, so try using
		 * gethostbyname() to look the name up.  Note that
		 * gethostbyname() blocks while looking the address up.
		 */

#if defined(OS_VXWORKS)
		host_entry = NULL;	/* FIXME */
#else
		host_entry = gethostbyname( hostname );
#endif

		if ( host_entry != NULL ) {

			/* Assign h_addr to a void pointer first in order
			 * to avoid the warning 'cast increases required
			 * alignment of target type' from GCC 2.8.1 under
			 * Solaris 2.6.
			 */

			void_ptr = host_entry->h_addr;

			in_addr_struct = (struct in_addr *) void_ptr;

			*inet_address = in_addr_struct->s_addr;
		} else {
#if defined(OS_WIN32)
			error_value = WSAGetLastError();
#elif defined(OS_VXWORKS)
			error_value = HOST_NOT_FOUND;	/* FIXME */
#else
			error_value = h_errno;
#endif
			switch( error_value ) {
			case HOST_NOT_FOUND:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
					"The host name '%s' was not found.",
					hostname );
				break;
			case NO_ADDRESS:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The requested domain name '%s' is valid, but does not have an IP address.",
					hostname );
				break;
			case NO_RECOVERY:
				return mx_error( MXE_FUNCTION_FAILED, fname,
				"The domain name '%s' does not exist.",
					hostname );
				break;
			case TRY_AGAIN:
				return mx_error( MXE_TRY_AGAIN, fname,
		"The domain name '%s' does not currently seem to exist, "
		"but this is likely to be a temporary condition, "
		"so try again later.", hostname );
				break;
			default:
				return mx_error( MXE_FUNCTION_FAILED, fname,
"An unrecognized error code %d was returned by the call gethostbyname('%s')",
					error_value, hostname );
				break;
			}
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_socket_send( MX_SOCKET mx_socket,
		void *message_buffer,
		size_t message_length_in_bytes,
		int flags )
{
	static const char fname[] = "mx_socket_send()";

	long bytes_left, bytes_sent;
	int saved_errno;
	char *ptr;

	bytes_left = (long) message_length_in_bytes;

	ptr = (char *) message_buffer;

	while( bytes_left > 0 ) {
		bytes_sent = send( mx_socket, ptr, bytes_left, 0 );

		switch( bytes_sent ) {
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			switch( saved_errno ) {
#ifdef OS_WIN32
			case WSAECONNRESET:
#else
			case ECONNRESET:
#endif
				return mx_error(
					MXE_NETWORK_CONNECTION_LOST, fname,
		"Network connection lost.  Errno = %d, error text = '%s'",
				saved_errno, mx_socket_strerror(saved_errno));
				break;
			default:
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Error sending to remote host.  Errno = %d, error text = '%s'",
				saved_errno, mx_socket_strerror(saved_errno));
				break;
			}
			break;
		default:
			bytes_left -= bytes_sent;
			ptr += bytes_sent;
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_socket_receive( MX_SOCKET mx_socket,
		void *message_buffer,
		size_t message_length_in_bytes,
		int flags )
{
	static const char fname[] = "mx_socket_receive()";

	long bytes_left, bytes_received;
	int i, saved_errno;
	char *ptr;

	bytes_left = (long) message_length_in_bytes;

	ptr = (char *) message_buffer;

	while( bytes_left > 0 ) {
		bytes_received = recv( mx_socket, ptr, bytes_left, 0 );

		switch( bytes_received ) {
		case 0:
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
				"Network connection closed unexpectedly." );
			break;
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			switch( saved_errno ) {
#ifdef OS_WIN32
			case WSAECONNRESET:
#else
			case ECONNRESET:
#endif
				return mx_error(
					MXE_NETWORK_CONNECTION_LOST, fname,
		"Network connection lost.  Errno = %d, error text = '%s'",
				saved_errno, mx_socket_strerror(saved_errno));

				break;
			default:
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Error receiving message body from remote host.  "
			"Errno = %d, error text = '%s'",
				saved_errno, mx_socket_strerror(saved_errno));

				break;
			}
			break;
		default:
			if ( flags & MXF_SOCKET_USE_LINE_TERMINATORS ) {

				/* See if a <CR> or <LF> character is present
				 * in the bytes we just read.  If one is,
				 * null it out and then return.
				 */

				for ( i = 0; i < bytes_received; i++ ) {
					if ( ( ptr[i] == MX_CR )
					  || ( ptr[i] == MX_LF ) )
					{
						ptr[i] = '\0';
						return MX_SUCCESSFUL_RESULT;
					}
				}
			}

			bytes_left -= bytes_received;
			ptr += bytes_received;
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_socket_putline( MX_SOCKET mx_socket, char *buffer )
{
	char terminators[] = "\015\012";
	mx_status_type mx_status;

	/* Send the buffer. */

	mx_status = mx_socket_send( mx_socket, buffer, strlen( buffer ), 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the line terminators. */

	mx_status = mx_socket_send( mx_socket, terminators,
					strlen( terminators ), 0 );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_socket_getline( MX_SOCKET mx_socket, char *buffer, size_t buffer_length )
{
	mx_status_type mx_status;

	mx_status = mx_socket_receive( mx_socket, buffer, buffer_length,
					MXF_SOCKET_USE_LINE_TERMINATORS );
	return mx_status;
}

#ifdef OS_WIN32

static void
mx_winsock_shutdown( void )
{
	int status, saved_errno;

	status = WSACleanup();

	if ( status != 0 ) {
		saved_errno = WSAGetLastError();

		fprintf(stderr,
		"WSACleanup() failed with error code %d, reason = '%s'\n",
			saved_errno, mx_socket_strerror( saved_errno ) );
	}
	return;
}

static mx_status_type
mx_winsock_initialize( void )
{
	static const char fname[] = "mx_winsock_initialize()";

	WORD version_requested;
	WSADATA winsock_data;
	int status;

	version_requested = MAKEWORD( 1, 1 );

	status = WSAStartup( version_requested, &winsock_data );

	if ( status != 0 ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Winsock initialization failed." );
	}

	if ( atexit( mx_winsock_shutdown ) != 0 ) {
		(void) WSACleanup();

		return mx_error( MXE_FUNCTION_FAILED, fname,
"Attempt to set up Winsock atexit() handler failed.  Winsock disabled." );
	}
	return MX_SUCCESSFUL_RESULT;
}

static char *
mx_winsock_strerror( int winsock_errno )
{
	int i;

	static struct {
		int error_code;
		char error_string[60];
	} winsock_error_messages[] = {

	{ WSAEINTR,		"Interrupted system call" },
	{ WSAEBADF,		"Bad file number" },
	{ WSAEACCES,		"Permission denied" },
	{ WSAEFAULT,		"Bad address" },
	{ WSAEINVAL,		"Invalid argument" },
	{ WSAEMFILE,		"Too many open files" },
	{ WSAEWOULDBLOCK,	"Operation would block" },
	{ WSAEINPROGRESS,	"Operation now in progress" },
	{ WSAEALREADY,		"Operation already in progress" },
	{ WSAENOTSOCK,		"Socket operation on nonsocket" },
	{ WSAEDESTADDRREQ,	"Destination address required" },
	{ WSAEMSGSIZE,		"Message too long" },
	{ WSAEPROTOTYPE,	"Protocol wrong type for socket" },
	{ WSAENOPROTOOPT,	"Bad protocol option" },
	{ WSAEPROTONOSUPPORT,	"Protocol not supported" },
	{ WSAESOCKTNOSUPPORT,	"Socket type not supported" },
	{ WSAEOPNOTSUPP,	"Operation not supported on socket" },
	{ WSAEPFNOSUPPORT,	"Protocol family not supported" },
	{ WSAEAFNOSUPPORT, "Address family not supported by protocol family" },
	{ WSAEADDRINUSE,	"Address already in use" },
	{ WSAEADDRNOTAVAIL,	"Cannot assign requested address" },
	{ WSAENETDOWN,		"Network is down" },
	{ WSAENETUNREACH,	"Network is unreachable" },
	{ WSAENETRESET,		"Net dropped connection or reset" },
	{ WSAECONNABORTED,	"Software caused connection abort" },
	{ WSAECONNRESET,	"Connection reset by peer" },
	{ WSAENOBUFS,		"No buffer space available" },
	{ WSAEISCONN,		"Socket is already connected" },
	{ WSAENOTCONN,		"Socket is not connected" },
	{ WSAESHUTDOWN,		"Cannot send after socket shutdown" },
	{ WSAETOOMANYREFS,	"Too many references, cannot splice" },
	{ WSAETIMEDOUT,		"Connection timed out" },
	{ WSAECONNREFUSED,	"Connection refused" },
	{ WSAELOOP,		"Too many levels of symbolic link" },
	{ WSAENAMETOOLONG,	"File name too long" },
	{ WSAEHOSTDOWN,		"Host is down" },
	{ WSAEHOSTUNREACH,	"No route to host" },
	{ WSAENOTEMPTY,		"Directory is not empty" },
	{ WSAEPROCLIM,		"Too many processes" },
	{ WSAEUSERS,		"Too many users" },
	{ WSAEDQUOT,		"Disk quota exceeded" },
	{ WSAESTALE,		"Stale NFS file handle" },
	{ WSASYSNOTREADY,	"Network subsystem is unavailable" },
	{ WSAVERNOTSUPPORTED,	"Winsock DLL version out of range" },
	{ WSANOTINITIALISED,	"Successful WSAStartup() not yet performed" },
	{ WSAEREMOTE,		"Too many levels of remote in path" },
	{ WSAHOST_NOT_FOUND,	"Host not found" },
	{ WSATRY_AGAIN,		"Nonauthoritative host not found" },
	{ WSANO_RECOVERY,	"Nonrecoverable error" },
	{ WSANO_DATA,	    "Valid name, no data record of requested type" },
	{ WSANO_ADDRESS,	"No address, look for MX record" }
	};

	static int num_winsock_error_messages = sizeof(winsock_error_messages)
				/ sizeof(winsock_error_messages[0]);

	static char unrecognized[] = "Unrecognized Winsock error code";

	for ( i = 0; i < num_winsock_error_messages; i++ ) {

		if ( winsock_errno == winsock_error_messages[i].error_code ) {

			return &(winsock_error_messages[i].error_string[0]);
		}
	}
	return &(unrecognized[0]);
}

#endif /* OS_WIN32 */

#ifdef OS_DJGPP

static mx_status_type
mx_watt32_initialize( void )
{
#if 0
	dbug_init();
#endif

	sock_init();

	return MX_SUCCESSFUL_RESULT;
}

#endif /* OS_DJGPP */

#endif /* USE_TCPIP */

