/*
 * Name:    ms_process_ccd.c
 *
 * Purpose: Functions used to process MX CCD record events.
 *
 * Author:  William Lavender
 *
 *------------------------------------------------------------------------
 *
 * Copyright 2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_ccd.h"
#include "mx_socket.h"

#include "ms_server.h"
#include "ms_process_ccd.h"

mx_status_type
mxsrv_setup_ccd_process_functions( MX_RECORD *record )
{
	const char fname[] = "mxsrv_setup_ccd_process_functions()";

	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	record_field_array = record->record_field_array;

	for ( i = 0; i < record->num_record_fields; i++ ) {

		record_field = &record_field_array[i];

		switch( record_field->label_value ) {
		case MXLV_CCD_DATA_FRAME_SIZE:
		case MXLV_CCD_BIN_SIZE:
		case MXLV_CCD_STATUS:
		case MXLV_CCD_START:
		case MXLV_CCD_STOP:
		case MXLV_CCD_READOUT:
		case MXLV_CCD_DEZINGER:
		case MXLV_CCD_CORRECT:
		case MXLV_CCD_WRITEFILE:
			record_field->process_function
					    = mxsrv_ccd_process_function;
			break;
		default:
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_ccd_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_ccd_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_CCD *ccd;
	mx_status_type status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;
	ccd = (MX_CCD *) record->record_class_struct;

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( record_field->label_value ) {
		case MXLV_CCD_DATA_FRAME_SIZE:
			status = mx_ccd_get_data_frame_size(record, NULL, NULL);
			break;
		case MXLV_CCD_BIN_SIZE:
			status = mx_ccd_get_bin_size( record, NULL, NULL );
			break;
		case MXLV_CCD_STATUS:
			status = mx_ccd_get_status( record, NULL );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	case MX_PROCESS_PUT:
		switch( record_field->label_value ) {
		case MXLV_CCD_BIN_SIZE:
			status = mx_ccd_set_bin_size( record,
					ccd->bin_size[0], ccd->bin_size[1] );
			break;
		case MXLV_CCD_START:
			status = mx_ccd_start( record );
			break;
		case MXLV_CCD_STOP:
			status = mx_ccd_stop( record );
			break;
		case MXLV_CCD_READOUT:
			status = mx_ccd_readout( record, ccd->frame_type );
			break;
		case MXLV_CCD_DEZINGER:
			status = mx_ccd_dezinger( record, ccd->frame_type );
			break;
		case MXLV_CCD_CORRECT:
			status = mx_ccd_correct( record );
			break;
		case MXLV_CCD_WRITEFILE:
			status = mx_ccd_writefile( record, ccd->writefile_name,
							ccd->frame_type );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return status;
}

